/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage4/control/client_auth.c,v 1.7 2001/05/08 15:19:44 koni Exp $ 
  $Author: koni $ $Date: 2001/05/08 15:19:44 $ $Name:  $
  $Revision: 1.7 $

*/
   
#ifdef WIN32
#include "slan_windows.h"
#else
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#endif

#include "slan_types.h"

#include "random_interface.h"
#include "sha.h"

#include "control_msg.h"
#include "control_tcp.h"

#include "slan_session.h"

#define CHAP_MESSAGE (0)

/* Nomber of microseconds until a various stage in the protocol timesout */
#define CHAP_TIME_MAX (3000000)

enum { CHAP_REQUIRED = 0, CHAP_NOTREQUIRED, CHAP_USERNAME, CHAP_NONCE, 
       CHAP_HASH, CHAP_ACCEPT, CHAP_REJECT, CHAP_ERROR };

enum { CHAP_OKAY = 0, CHAP_ESHORTNONCE, CHAP_EPASSWORD, CHAP_ETIMEOUT,
       CHAP_NERROR };

static void chap_compute_hash(unsigned char *hash_value, 
			      slan_session_t *session_obj, 
			      unsigned char *nonce,
			      int nonce_length) {
  unsigned char *hash_document;
  int hash_document_length, offset;

  /* We use twice the password length here so we can have some padding at the
     end of the begining of the document -- the amount of which we'll make
     dependant on the size of the password, for an extra little security */
  hash_document_length = session_obj->username_length +
    session_obj->password_length*2 + nonce_length;

  hash_document = slan_malloc(hash_document_length);

  memset(hash_document,0x5c,hash_document_length);
  /* Allow a password (size) dependant amount of pre-padding */
  offset = session_obj->password_length/2;
  
  memcpy(hash_document + offset, session_obj->username, 
	 session_obj->username_length);
  offset += session_obj->username_length;

  memcpy(hash_document + offset, session_obj->password, 
	 session_obj->password_length);
  offset += session_obj->password_length;

  memcpy(hash_document + offset, nonce, nonce_length);
  
  sha_digest(hash_document, hash_document_length, hash_value);

  /* Do it again -- hash functions may leak some information that allow
     narrowing the search space for an offline brute force attack on the
     password. Hashing the result of the first hash should be a secure enough
     construction. */
  memset(hash_document, 0xa3, hash_document_length);
  memcpy(hash_document + session_obj->username_length, hash_value, 20);
  sha_digest(hash_document, hash_document_length, hash_value);
  
  slan_memfree(hash_document);
}

static int chap_get_elapsed_time(struct timeval tv) {
  struct timeval tv1;
  int rval;
 
  gettimeofday(&tv1, NULL);
  rval = (tv1.tv_sec - tv.tv_sec) * 1000000;
  rval += (tv1.tv_usec - tv.tv_usec);
  
  return rval;
}

/* Temporary */
static void server_get_password(slan_session_t *session_obj) {
  static unsigned char default_password[] = "yomama";

  session_obj->password = strdup(default_password);
  session_obj->password_length = strlen(default_password);
}

static int server_send_accept(slan_session_t *session_obj) {
  control_msg_t *msg;

  msg = new_control_msg(CLIENT_AUTH, CHAP_ACCEPT, session_obj->control_fd);
  control_msg_send(msg);
  control_msg_destruct(msg);

  return CHAP_OKAY;
}

static int server_send_reject(slan_session_t *session_obj) {
  control_msg_t *msg;

  msg = new_control_msg(CLIENT_AUTH, CHAP_REJECT, session_obj->control_fd);
  control_msg_send(msg);
  control_msg_destruct(msg);

  return CHAP_EPASSWORD;
}

static int server_listen_hash(slan_session_t *session_obj, 
			      unsigned char *nonce) {
  struct timeval tv;
  control_msg_t *msg;
  int type, subtype;
  unsigned char hash_value[20];
  unsigned char *client_hash;
  int client_hash_length;

  chap_compute_hash(hash_value, session_obj, nonce, 12);

  gettimeofday(&tv, NULL);
  do {
    msg = control_msg_receive(session_obj->control_fd, &type, &subtype,
			      CHAP_TIME_MAX);
    if (msg!=NULL) {
      if (type == CLIENT_AUTH && subtype == CHAP_ERROR) {
	/* Currently, the only error we are expecting from the client is a
	   short nonce value error. */
	control_msg_destruct(msg);
	return CHAP_ESHORTNONCE;
      } else if (type == CLIENT_AUTH && subtype == CHAP_HASH) {
	msg_search_tlv(msg, CHAP_HASH, &client_hash_length, &client_hash);
	control_msg_destruct(msg);
	if (client_hash_length != 20 || memcmp(client_hash, hash_value, 20)) {
	  if (client_hash!=NULL) slan_memfree(client_hash);
	  return server_send_reject(session_obj);
	}
	slan_memfree(client_hash);
	return server_send_accept(session_obj);
      }
    }
    control_msg_destruct(msg);
  }while (chap_get_elapsed_time(tv) < CHAP_TIME_MAX);
  return CHAP_ETIMEOUT;
}


static int server_send_nonce(slan_session_t *session_obj) {
  control_msg_t *msg;
  unsigned char nonce[12];
  
  random_byte_string(nonce, 12);
  msg = new_control_msg(CLIENT_AUTH, CHAP_NONCE, session_obj->control_fd);
  msg_append_tlv(msg, CHAP_NONCE, 12, nonce);
  control_msg_send(msg);
  control_msg_destruct(msg);
  
  return server_listen_hash(session_obj, nonce);
}

static int server_listen_username(slan_session_t *session_obj) {
  struct timeval tv;
  control_msg_t *msg;
  int type, subtype;

  gettimeofday(&tv, NULL);
  do {
    msg = control_msg_receive(session_obj->control_fd, &type, &subtype, 
			      CHAP_TIME_MAX);
    if (msg != NULL) {
      if (type == CLIENT_AUTH && subtype == CHAP_USERNAME) {
	msg_search_tlv(msg, CHAP_USERNAME, &session_obj->username_length,
		       &session_obj->username);
	control_msg_destruct(msg);
	if (session_obj->username == NULL) continue;
	server_get_password(session_obj);
	return server_send_nonce(session_obj);
      }
      control_msg_destruct(msg);
    }
  } while (chap_get_elapsed_time(tv) < CHAP_TIME_MAX);
  return CHAP_ETIMEOUT;
}


int server_initiate_client_auth(slan_session_t *session_obj) {
  control_msg_t *msg;
  
  msg = new_control_msg(CLIENT_AUTH, CHAP_REQUIRED, session_obj->control_fd);
  control_msg_send(msg);
  control_msg_destruct(msg);

  return server_listen_username(session_obj);
}


static int client_listen_accept(slan_session_t *session_obj) {
  struct timeval tv;
  control_msg_t *msg;
  int type, subtype;

  gettimeofday(&tv, NULL);
  do {
    msg = control_msg_receive(session_obj->control_fd, &type, &subtype, 
			      CHAP_TIME_MAX);
    if (msg != NULL) {
      if (type == CLIENT_AUTH && subtype == CHAP_ACCEPT) {
	control_msg_destruct(msg);
	return CHAP_OKAY;
      } else if (type == CLIENT_AUTH && subtype == CHAP_REJECT) {
	control_msg_destruct(msg);
	return CHAP_EPASSWORD;
      }
      control_msg_destruct(msg);
    }
  } while (chap_get_elapsed_time(tv) < CHAP_TIME_MAX);
  return CHAP_ETIMEOUT;
}

static int client_send_hash(slan_session_t *session_obj, unsigned char *nonce, 
			    int nonce_length) {
  control_msg_t *msg;
  unsigned char hash_value[20];
  
  
  /* Expect at least 8 bytes of nonce. If we get a short nonce, we abort the
     protocol. */
  if (nonce_length < 8) {
    int error_code;
    msg = new_control_msg(CLIENT_AUTH, CHAP_ERROR, session_obj->control_fd);
    error_code = CHAP_ESHORTNONCE;
    msg_append_tlv(msg, CHAP_ERROR, sizeof(error_code), 
		   (unsigned char *) &error_code);
    control_msg_send(msg);
    control_msg_destruct(msg);
    return CHAP_ESHORTNONCE;
  }

  chap_compute_hash(hash_value, session_obj, nonce, nonce_length);

  msg = new_control_msg(CLIENT_AUTH, CHAP_HASH, session_obj->control_fd);
  msg_append_tlv(msg, CHAP_HASH, 20, hash_value);
  control_msg_send(msg);
  control_msg_destruct(msg);

  return client_listen_accept(session_obj);
}

static int client_listen_nonce(slan_session_t *session_obj) {
  struct timeval tv;
  control_msg_t *msg;
  int type, subtype;
  unsigned char *nonce;
  int nonce_length;

  gettimeofday(&tv, NULL);
  do {
    msg = control_msg_receive(session_obj->control_fd, &type, &subtype, 
			      CHAP_TIME_MAX);
    if (msg != NULL) {
      if (type == CLIENT_AUTH && subtype == CHAP_NONCE) {
	msg_search_tlv(msg, CHAP_NONCE, &nonce_length, &nonce);
	control_msg_destruct(msg);
	if (nonce!=NULL) {
	  int rval;
	  rval = client_send_hash(session_obj, nonce, nonce_length);
	  slan_memfree(nonce);
	  return rval;
	}
      }
      control_msg_destruct(msg);
    }
  } while (chap_get_elapsed_time(tv) < CHAP_TIME_MAX);
  return CHAP_ETIMEOUT;
}



static int client_send_username(slan_session_t *session_obj) {
  control_msg_t *msg;

  msg = new_control_msg(CLIENT_AUTH, CHAP_USERNAME, session_obj->control_fd);
  msg_append_tlv(msg, CHAP_USERNAME, session_obj->username_length,
		 session_obj->username); 
  control_msg_send(msg);
  control_msg_destruct(msg);
  
  return client_listen_nonce(session_obj);
}
  

int client_authenticate_listen(slan_session_t *session_obj) {
  struct timeval tv;
  control_msg_t *msg;
  int type, subtype;

  gettimeofday(&tv, NULL);
  do {
    msg = control_msg_receive(session_obj->control_fd, &type, &subtype, 
			      CHAP_TIME_MAX);
    if (msg != NULL) {
      if (type == CLIENT_AUTH && subtype == CHAP_REQUIRED) {
	control_msg_destruct(msg);
	return client_send_username(session_obj);
      } 
      else if (type == CLIENT_AUTH && subtype == CHAP_NOTREQUIRED) {
	control_msg_destruct(msg);
	return CHAP_OKAY;
      }
      control_msg_destruct(msg);
    }
  } while (chap_get_elapsed_time(tv) < CHAP_TIME_MAX);
  return CHAP_ETIMEOUT;
}

