/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage4/control/control_msg.c,v 1.11 2001/06/25 02:12:20 koni Exp $ 
  $Author: koni $ $Date: 2001/06/25 02:12:20 $ $Name:  $
  $Revision: 1.11 $

*/
   
#ifndef WIN32
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <syslog.h>
#else
#include "slan_windows.h"
#endif

#include "slan_types.h"
#include "slan_alloc.h"

#include "slan_session.h"
#include "sym_cipher.h"
#include "mac.h"
#include "control_msg.h"

/* Byte order conversion functions -- htonl() and ntohl() make me nervous
   if the actual function is for some reason implemented as a macro and it
   is called with something like htonl(x++). We can get away with using a
   macro to alias ntoh32(x) to htonl32() because the actaul operation is done
   by a real function which won't evaluate 'x++' serveral times.

   This function simply swaps the order of the bytes in a 32 bit
   word. Necessary on stupid little-endian architectures like the ubiquitous
   x86 */
#define ntoh32(x) hton32((x));
static INLINE_FUNC u_int32_t hton32(u_int32_t x) {
#if __BYTE_ORDER == __LITTLE_ENDIAN  
  /* This is not going to sign-extend on the right shift, right? */
  x = (x << 16) | (x >> 16);
  x = ((x & 0xFF00FF00) >> 8) | ((x & 0x00FF00FF) << 8);
#endif
  return x;
}

/* Allocate and setup a new control message structure, reserving some room for 
   header information where security/integrity stuff will be needed and some
   protocol type/subtype information */
control_msg_t *new_control_msg(u_int32_t protocol, u_int32_t type, 
			  
     slan_session_t *session_obj) {
  control_msg_t *msg;
  u_int32_t x;
  int header_reserve;
  unsigned char *p;

  /* Allocate memory for a new control message, return NULL if allocation
     fails. */
  msg = slan_calloc(1, sizeof(control_msg_t));
  if (msg == NULL) return NULL;

  msg->buf = slan_calloc(DEFAULT_MSG_SIZE, 1);
  if (msg->buf == NULL) {
    slan_memfree(msg);
    return NULL;
  }

  /* This is important -- when we send the message we'll need to have the
     pointers to the cipher and mac objects to sign/encrypt the message */
  msg->session_obj = session_obj;

  msg->allocated_size = DEFAULT_MSG_SIZE;
  
  /* Save room for the following:

     Control Message Constant (Magic Number) 4 bytes
     Message Total Length                    4 bytes
     
     The message authentication code, whatever size that happens to be.
     Control Message Sequence Number         4 bytes
     Protocol Type                           4 bytes
     Message Type (within Protocol)          4 bytes */
  header_reserve = session_obj->mac->hash_size + 5*sizeof(u_int32_t); 
  msg->head = msg->buf + header_reserve;
  msg->length = header_reserve;
  msg->tail = msg->head;

  /* Set some header information */
  x = hton32(CONTROL_MSG_TYPE);
  memcpy(msg->buf, &x, sizeof(u_int32_t));

  /* Skip 8 bytes for control constant and message total length, then
     the variable size MAC, then 4 more bytes for the sequence number */
  p = msg->buf + 8 + session_obj->mac->hash_size + 4;

  /* Now, we'll set the protocol and type */
  protocol = hton32(protocol);
  type = hton32(type);
  memcpy(p, (unsigned char *) &protocol, sizeof(u_int32_t));
  memcpy(p + 4, (unsigned char *) &type, sizeof(u_int32_t));

  return msg;
}

/* Get rid of a control message when we are done with it. In principal, this
   should probably be called by the send function after sucessfully sending
   the message but currently the control code call the function directly -- a
   stupid idea if we continue to use TCP as the transport protocol since in
   theory the message need only be sent once. If we switch to UDP (a potential 
   option) then we may have to deal with retranmission at higher layers. */
void control_msg_destruct(control_msg_t *msg) {
  
  if (msg == NULL) return;
  if (msg->buf != NULL) slan_memfree(msg->buf);
  slan_memfree(msg);
}

/* This function is called to stuff some data into a control message using TLV 
   (type length value) encoding. Hence, we need a type and length with the
   data. */
void msg_append_tlv(control_msg_t *msg, u_int32_t type, int length,
		    unsigned char *value) {
  u_int32_t l;
  
  if ((msg->length + length + sizeof(u_int32_t)*2)> msg->allocated_size) {
    unsigned char *tmp;
    tmp = slan_realloc(msg->buf, msg->allocated_size+length+sizeof(u_int32_t)*2);
    if (tmp == NULL) {
      /* Silently fail -- this will cause the receiver probably to balk at the 
	 message and abort the protocol, but if we're out of memory, we're out 
	 of memory. */
      return;
    }
    msg->buf = tmp;
    msg->allocated_size += length;
  }

  type = hton32(type);
  memcpy(msg->tail, (unsigned char *) &type, sizeof(u_int32_t));
  msg->tail += sizeof(u_int32_t);
  msg->length += sizeof(u_int32_t);
  
  l = hton32(((u_int32_t) length));
  memcpy(msg->tail, (unsigned char *) &l, sizeof(u_int32_t));
  msg->tail += sizeof(u_int32_t);
  msg->length += sizeof(u_int32_t);

  memcpy(msg->tail,value,length);
  msg->tail += length;
  msg->length += length;
}

/* Important! We can seriously run aground if we blindly expect TLVs to appear 
   where they should and have correct length codes. Otherwise, a rogue packet
   could trick us into a segmentation fault, bus error, or worse, heap
   smashing buffer overflow. 

   Thus, before copying any data out of the message into a variable or
   allocating memory to hold the VALUE portion of TLV, we first check the
   bounds to see if the data exists. This prevents us from scaning for TLVs
   off in the middle of nowhere in the memory. */
static int msg_safe_pullvalue(control_msg_t *msg, unsigned char *p, int
			      length, void *value) {
  
  if (p + length > msg->tail) return 1;
  else {
    memcpy(value, p, length);
  }
  return 0;
}

static int msg_safe_allocpullvalue(control_msg_t *msg, unsigned char *p, int
				   length, unsigned char **value) {

  if (p + length > msg->tail) return 1;
  else {
    *value = slan_malloc(length + 1);
    if (*value == NULL) return 1;
    memcpy(*value, p, length);
    /* Add terminating NULL character for convienence -- in case this value is
       ever used in string functions */
    (*value)[length] = 0;
  }

  return 0;
}


/* This function does the dirty work of finding a TLV encoded value in 
   a control_msg_t structure. It is inefficient in the message
   contains many TLVs, because it scans possibly the whole message
   each time its called. Types and lengths are expected to be in
   big-endian format, for consistany, so little endian manchines need
   to rearrange the bytes before interpreting the types and lengths. 
   
   If a message contains two TLVs with the same type that we are
   looking for, the first one will be used. */
void msg_search_tlv(control_msg_t *msg, u_int32_t type,
		    int *length, unsigned char **value) {
  u_int32_t l, scan_type;
  unsigned char *p;

  /* Initialize these first, so that if an error occurs due to a bogus
     message, we can just bail out right at that point */
  *value = NULL; *length = 0;

  p = msg->head;
  type = hton32(type);
  while(p < msg->tail) {
    /* Expecting the 4-byte TYPE code of a TLV, pull it from the message */
    if (msg_safe_pullvalue(msg, p, sizeof(u_int32_t), &scan_type)) return;
    p += sizeof(u_int32_t);

    /* Next we expect the length of the _VALUE_ part of the TLV, in 4 bytes */
    if (msg_safe_pullvalue(msg, p, sizeof(u_int32_t), &l)) return;
    l = ntoh32(l);
    p += sizeof(u_int32_t);

    if (scan_type == type) {
      /* Search HIT */
      *length = l;
      if (msg_safe_allocpullvalue(msg, p, *length, value)) {
	*length = 0;
      }
      return;
    } else {
      /* Skip over a different TLV */
      p += l;
    }
  }

  /* TLV was not found if here */
}

void control_msg_send(control_msg_t *msg) {
  int send_len, mac_length;
  u_int32_t sequence_no, length;
  u_int32_t control_constant;

  mac_length = msg->session_obj->mac->hash_size;

  /* Debug shit */
#ifdef DEBUG
  fprintf(stderr,"Sending message of type %0X\n",
	  *((u_int32_t *) (msg->buf + 12 + mac_length)));
  
  fprintf(stderr,"subtype %0X\n",
	  *((u_int32_t *) (msg->buf + 16 + mac_length)));
  fprintf(stderr,"length %0X\n",
	  *((u_int32_t *) (msg->buf + 4)));
#endif

  /* Place a sequence number of the packet, which will be included in the MAC
     computation. This is to frustrate an attacker attempting to replay old
     valid messages to frustrate the control protocol. We need to transmit in
     network byte order so any host can interpret it back into native
     format. As well, I think htonl() may sometimes be implemented as a macro, 
     so we'll do the increment afterwards. Since the MAC may be some odd
     number of bytes that is not divisible by the machine word size, we use
     memcpy to place the sequence number so that a bus error does not occur. */
  sequence_no = hton32(msg->session_obj->control_seq_out);
  /* Mutex Lock ? */
  msg->session_obj->control_seq_out++;
  
  memcpy(msg->buf + mac_length + 8, (unsigned char *) &sequence_no, 
	 sizeof(u_int32_t)); 

  /* Now, we sign the message using the MAC configured for this session and
     place it at the begining of the data buffer. We cannot use the nocopy
     sign method because we don't have adequate header/trailer room reserved.*/
  msg->session_obj->mac->message_sign(msg->session_obj->mac, 
				      msg->buf + mac_length + 8, 
				      msg->length - mac_length - 8, 
				      msg->buf + 8);

  /* Finally, encrypt the message for privacy, using the configured symmetric
     cipher for this session. */
  msg->session_obj->cipher->cbc_encrypt(msg->session_obj->cipher, 
					msg->buf + 8, msg->length - 8); 

  /* Plug the control message constant and length into the first eight bytes
     of this message so the receiving end recognizes it properly */
  control_constant = hton32(CONTROL_MSG_TYPE);
  memcpy(msg->buf, (unsigned char *) &control_constant, sizeof(u_int32_t));

  length = hton32(msg->length);;
  memcpy(msg->buf + 4, (unsigned char *) &length, sizeof(u_int32_t));

  /* Send the message on its way */
  
  send_len = send(msg->session_obj->control_fd, msg->buf, msg->length, 
#ifdef WIN32
		  0);
#else
		  MSG_NOSIGNAL);
#endif

  /* Error control here  -- TO BE HANDLED BETTER */
  if (send_len < msg->length) {
    perror("control_msg: Error sending message: ");
  }
}

int control_get_elapsed_time(struct timeval tv) {
  struct timeval tv1;
  int rval;
 
  gettimeofday(&tv1, NULL);
  rval = (tv1.tv_sec - tv.tv_sec) * 1000000;
  rval += (tv1.tv_usec - tv.tv_usec);
  
  return rval;
}

