/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage4/control/control_tcp.c,v 1.14 2001/07/06 17:30:12 lucas Exp $ 
  $Author: lucas $ $Date: 2001/07/06 17:30:12 $ $Name:  $
  $Revision: 1.14 $

*/
   
#ifdef WIN32
#include "slan_windows.h"
#else
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <syslog.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif

#include "slan_types.h"
#include "slan_alloc.h"

#include "control_tcp.h"

/* This function opens, binds, and sets a TCP/IP socket to listen for
   incomming connections. */
SOCKET server_open_socket(int port) {
  struct sockaddr_in address;
  SOCKET tcp_socket;

  tcp_socket = socket(PF_INET, SOCK_STREAM, 0);
  if (tcp_socket < 0) {
    perror("Can not start SLAN: Can not open TCP socket: ");
#ifndef WIN32
    syslog(LOG_ERR,"Can not start SLAN: Can not open TCP socket (%m)\n");
#endif
    exit(-1);
  }
  
  memset(&address, 0, sizeof(struct sockaddr_in));

  address.sin_family = AF_INET;
  address.sin_addr.s_addr = htonl(INADDR_ANY);
  address.sin_port = htons(port);

  if (bind(tcp_socket, (struct sockaddr *)&address, sizeof(struct sockaddr_in))) {
    perror("Can not start SLAN: Can not bind TCP socket: ");
#ifndef WIN32
    syslog(LOG_CRIT,"Can not start SLAN: Can not bind TCP socket (%m)\n");
#endif

    exit(-1);
  }

  if (listen(tcp_socket, 100)) {
    perror("Can not start SLAN: Cannot make TCP socket listen for "
	   "connections: ");
#ifndef WIN32
    syslog(LOG_CRIT,"Can not start SLAN: Can not make TCP socket listen for "
	   "connections (%m)\n");
#endif
    exit(-1);
  }
  return tcp_socket;
}

/* This function opens a socket for a client, and attempts to connect
   to the server specified by name or ip_address "server_name" */
SOCKET client_open_socket(unsigned char *server_name, 
			  unsigned int *server_binaddr,
			  int port) {
  SOCKET tcp_socket;
  struct sockaddr_in server_addr;
  struct hostent *hostname_hent;

  tcp_socket = socket(PF_INET, SOCK_STREAM, 0);
  if (tcp_socket < 0) {
    perror("client: Cannot open TCP socket: ");
#ifndef WIN32
    syslog(LOG_CRIT,"Can not start SLAN: Can not open TCP socket (%m)\n");
#endif
    LEAVE_THREAD(-1);
  }

  memset(&server_addr, 0, sizeof(struct sockaddr_in));

  /* This could almost be a bug, if the address passed to us is 255.255.255.255.. ;) */  

  if((server_addr.sin_addr.s_addr = inet_addr(server_name)) == INADDR_NONE)
  {

    hostname_hent = gethostbyname(server_name);
    if (hostname_hent == NULL) {
#ifndef WIN32
      fprintf(stderr,"client: Could not resolve server hostname %s: %s\n",
              server_name, hstrerror(h_errno));
      syslog(LOG_ERR,"Cannot start SLAN. Could not resolve server hostname "
	     "%s (%m)\n",server_name);
#endif
      LEAVE_THREAD(-1);
    }

    memcpy(&server_addr.sin_addr.s_addr,hostname_hent->h_addr_list[0],
  	   hostname_hent->h_length);
  }

  *server_binaddr = server_addr.sin_addr.s_addr;

  server_addr.sin_port = htons(port);
  server_addr.sin_family = AF_INET;

  if (connect(tcp_socket, (struct sockaddr *) &server_addr,sizeof(struct sockaddr_in))) {
    perror("client: Could not connect to server: ");
    LEAVE_THREAD(-1);
  }

  return tcp_socket;
}

/* Accepts a connection from a client, returning a new socket
   descriptor for that connection */
SOCKET server_accept(SOCKET fd, struct sockaddr_in *address, socklen_t *addrlen) {
  SOCKET connection_socket;

  *addrlen = sizeof(struct sockaddr_in);
  connection_socket = accept(fd, (struct sockaddr *)address, addrlen);
  if (connection_socket < 0) {
    perror("server: Error attempting to accept connection: ");
    exit(-1);
  }
  return connection_socket;
}


/* This message peeks at data from the tcp receive window, performing a linear 
   search until a control constant is found, discarding any bytes that do not
   have the constant */
static int tcp_buffer_bleed(int fd) {
  unsigned char junk_buf[4096];
  int bytes_read, i;
  unsigned char target[4];
  u_int32_t t;

  /* Make a copy of the control constant to search for */
  t = CONTROL_MSG_TYPE;
  t = htonl(t);
  memcpy(target, &t, 4);

  /* Get some data to poke through */
  bytes_read = recv(fd, junk_buf, 4096, MSG_PEEK);
  if (bytes_read <= 0) return 0;

  for(i=0;i<bytes_read;i++) {
    /* Ouch... linear search */
    if (memcmp(junk_buf + i, target, 4)==0) {
      /* Pull the garbage out of the buffer, plus the control constant */
      bytes_read = recv(fd, junk_buf, i + 4, 0);
      return bytes_read;
    }
  }

  /* Get rid of all that garbage */
  recv(fd, junk_buf, bytes_read, 0);

  /* Nope, no needle in the haystack */
  return 0;
}

#ifdef WIN32
#define control_recv_func slan_w32_recvall
#else
#define control_recv_func recv
#endif

/* This function reads data from the non-blocking socket, hoping to get enough 
   to satisfy the request, or none at all. If no data is ready within the
   timeout period, the select call will stop the operation and return. If some 
   data is available, but not all of it, we'll wait for the rest possibly much 
   longer than the timeout value (if it keeps comming in little pieces). This
   is desireable because if we give up on part of a control message, most
   likely the setup stage of the protocol will fail, and that is annoying. */
static int tcp_timed_read(SOCKET fd, unsigned char *buf, int length, 
			  int microseconds) {

#ifndef WIN32
  fd_set readfds;
  struct timeval tv;
#endif
  int bytes_read;
  
#ifndef WIN32
  /* First, setup a select() statement to poll for the data comming in */
  FD_ZERO(&readfds);
  FD_SET(fd,&readfds);
  
  tv.tv_sec = microseconds/1000000;
  tv.tv_usec = microseconds % 1000000;

  select(fd+1, &readfds, NULL, NULL, &tv);
  if (!FD_ISSET(fd,&readfds)) {
    /* No bytes read */
    return 0;
  }
#endif

  /* Socket should be marked non-blocking, so this should return immediately */
  bytes_read = recv(fd, buf, length, 0);

  /* Good, this is what we hope usually happens */
  if (bytes_read == length) return bytes_read;

  /* Keep trying to get that data with an exponential decay on the
     timeout. Once we are down to 1/10 of a second, we'll stop bothering with
     running this loop as it is not really worth it. The reason we exceed our
     time limit (possibly) to get the rest of the data is that once we've read 
     part of the request, if it is valid control message, it won't be able to
     processed unless its complete. Thus, if we fail to get all of it (and
     really its been delayed slightly), the caller of this function will be 
     forced to drop the message, and most likely the control protocol using
     that message will fail (due to timeouts) at both ends. So we make an
     effort to get a complete message, but still gaurantee that we will not
     get stuck here forever. */
  if (microseconds > 100000) {
    return bytes_read + tcp_timed_read(fd, buf + bytes_read, length-bytes_read,
				       microseconds/2);
  } 

  /* Oh well, if that data is comming, we're all fucked up now because we
       are not waiting any longer! */
  return bytes_read;  
}

/* This message attempts to receive and verify a control message. The argument 
   <microseconds> provides a timeout period -- we may end up waiting much
   longer than that (see function above) trying to finish a read -- but we
   will eventually fail as the retry works as a finite geometric sum.

   Failing to receive a control message may cause a higher level timeouts and
   cause a stage to fail, which may prevent session setup. Probably, timeouts
   should be set higher there so that we can tolerate a reasonable amount of
   smokescreen sent to us by some attacker, and still setup the connection as
   long as the legitimate party can still get a word in edgewise. */
control_msg_t *control_msg_receive(slan_session_t *session_obj, int *r_type, 
				   int *r_subtype, int microseconds) {
  unsigned char *buf, *p;
  unsigned char local_hash[64];
  int length, mac_length;
  u_int32_t msg_length, t, type, subtype;
  u_int32_t control_constant;
  u_int32_t sequence_no;
  control_msg_t *msg;
  int fd;

  fd = session_obj->control_fd;
  msg = slan_calloc(1, sizeof(control_msg_t));
  buf = slan_malloc(10240);
  if (buf == NULL) return NULL;
  
  length = tcp_timed_read(fd, buf, sizeof(u_int32_t), microseconds);
  if (length != sizeof(u_int32_t)) {
#ifndef WIN32
    syslog(LOG_ERR,"Error -- timeout waiting for control message.\n");
#endif
    goto fail;
  }

  control_constant = htonl(CONTROL_MSG_TYPE);
  memcpy((unsigned char *) &t, buf, sizeof(u_int32_t));

  if (control_constant != t) {
    if (tcp_buffer_bleed(fd)) {
      memcpy(buf, &control_constant, 4);
    } else {
#ifndef WIN32
      syslog(LOG_ERR,"Error -- control constant not found where expected.\n");
#endif
      goto fail;
    }
  }

  length = tcp_timed_read(fd, buf + 4, 4, microseconds);
  if (length != 4) {
#ifndef WIN32
    syslog(LOG_ERR,"Error -- unable to read length of control message.\n");
#endif
    goto fail;
  }
  
  memcpy((unsigned char *) &t, buf + 4, 4);
  msg_length = ntohl(t);
  if (msg_length + 8 > CONTROL_MSG_MAX) {
    /* Oh pain in the ass.... somebody must have sent a really long message or 
       somehow we are reading something that is not a control message. */
#ifndef WIN32
    syslog(LOG_ERR,"Error -- control message is too long.\n");
#endif
    goto fail;
  }

  length = tcp_timed_read(fd, buf + 8, (msg_length - 8), microseconds);
  if (length != msg_length - 8) {
#ifndef WIN32
    syslog(LOG_ERR,"Error -- unable to read full control message.\n");
#endif
    goto fail;
  }

  msg->session_obj = session_obj;
  msg->length = msg_length;
  msg->allocated_size = msg_length;
  msg->buf = buf;
  msg->tail = msg->buf + msg_length;
  msg->head = msg->buf + 5*sizeof(u_int32_t) + session_obj->mac->hash_size;

  /* Decrypt the control message */
  session_obj->cipher->cbc_decrypt(session_obj->cipher, msg->buf + 8, 
				   msg->length - 8);

  mac_length = session_obj->mac->hash_size;
  /* Prevent hell from breaking loose -- if this is greater than 64 bytes than 
     something is wrong. The next step will certainly fail. Lets just try to
     stop the whole server from coming down. */
  if (mac_length > 64) mac_length = 64;
  session_obj->mac->message_sign(session_obj->mac, msg->buf + mac_length + 8,
				 msg->length - 8 - mac_length, local_hash);
  if (memcmp(msg->buf + 8, local_hash, mac_length)) {
    /* Invalid MAC */
#ifndef WIN32
    syslog(LOG_WARNING,"Rejecting control message with invalid MAC.\n");
#endif
    goto fail;
  }

  /* Check the sequence number. This is TCP, so we tolerate no losses at
     all. The sequence number must match exactly. */
  memcpy((unsigned char *) &sequence_no, msg->buf + 8 + mac_length, 
	 sizeof(u_int32_t)); 
  sequence_no = ntohl(sequence_no);
  if ( sequence_no != session_obj->control_seq_in ) {
    /* Message out of sequence -- possible replay attack */
#ifndef WIN32
    syslog(LOG_WARNING,"Rejecting out-of-sequence control message.\n");
#endif
    goto fail;
  }
  session_obj->control_seq_in++;

  /* Get at the start of the message, beyond the CONTROL CONSTANT, total
     length, MAC, and sequence number */
  p = buf + 12 + mac_length;
  memcpy(&type, p, sizeof(u_int32_t));
  memcpy(&subtype, p + 4, sizeof(u_int32_t));

  *r_type = ntohl(type);
  *r_subtype = ntohl(subtype);

  /* Debug shit */
#ifdef DEBUG
  fprintf(stderr,"Received message of type %0X\n", *r_type);  
  fprintf(stderr,"subtype %0X\n", *r_subtype);
  fprintf(stderr,"length %0X\n", msg_length);
#endif
  return msg;


 fail:
  if (msg) free(msg);
  if (buf) free(buf);
  return NULL;
}




  
  

