/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage4/control/dh_keyexchange.c,v 1.14 2001/06/25 02:12:21 koni Exp $ 
  $Author: koni $ $Date: 2001/06/25 02:12:21 $ $Name:  $
  $Revision: 1.14 $

*/
   
#ifndef WIN32
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <syslog.h>
#else
#include "slan_windows.h"
#endif

#include "slan_types.h"
#include "slan_alloc.h"
#include "slan_config.h"

#ifdef HAVE_GMP_H
#include <gmp.h>
#else
#include "slan_gmp3.h"
#endif

#include "slan_session.h"

#include "dh_keyexchange.h"
#include "random_interface.h"
#include "control_tcp.h"
#include "sha.h"

extern slan_conf_t *slan_global_config;

enum { DH_INVITE = 0, DH_REQUEST, DH_DENY, DH_RESPONSE, DH_PUBLIC, DH_KEY,
       DH_KEYACK, DH_ABORT, DH_NUM_MSGTYPE };

enum { DH_GENERATOR = 0, DH_MODULUS, DH_PUBLIC_VALUE, DH_KEYHASH, 
       DH_ERROR };

enum { DH_FINISHED = 0, DH_EDENY, DH_EBADGN, DH_EBADPUBLIC,
       DH_EBADHASH, DH_ETIMEOUT, DH_NUMERROR};

static char *dh_error[DH_NUMERROR] = 
{ "(No Error) Diffie-Hellman key exchange complete",
  "Diffie Hellman key exchange request was refused by server",
  "Server sent invalid values for Diffie-Hellman key exchange",
  "Improper public value was received.",
  "DH key acknowledgement handshake with peer failed",
  "Protocol Timeout: Appropriate response from peer was not received" };

/* This is the amount of time in mircoseconds before the protocol will 
   timeout (and abort) waiting for a response from the peer */
#define DH_TIME_MAX (7000000)

typedef struct {
  slan_session_t *session_obj;
  mpz_t modulus_mpz;
  mpz_t generator_mpz;
  mpz_t public_mpz;
  mpz_t private_mpz;
  mpz_t key_mpz;
} dh_obj_t;

static dh_obj_t *new_dh_obj(slan_session_t *session_obj) {
  dh_obj_t *obj;
  
  obj = slan_malloc(sizeof(dh_obj_t));
  if (obj == NULL) return NULL;

  obj->session_obj = session_obj;
  mpz_init(obj->modulus_mpz);
  mpz_init(obj->generator_mpz);
  mpz_init(obj->public_mpz);
  mpz_init(obj->private_mpz);
  mpz_init(obj->key_mpz);
  
  return obj;
}

static void dh_obj_destruct(dh_obj_t *obj) {
  mpz_clear(obj->modulus_mpz);
  mpz_clear(obj->generator_mpz);
  mpz_clear(obj->public_mpz);
  mpz_clear(obj->private_mpz);
  mpz_clear(obj->key_mpz);

  slan_memfree(obj);
}
  

/* This is a utility function, to return the number of microseconds that have
   elapsed since the time recored in parameter <tv>. It is used to determine
   if a particular stage of the Diffie-Hellman keyexchange has timed out. By
   checking for timeouts when waiting for responses to arrive, we can
   guarantee finite progress. */
static int dh_get_elapsed_time(struct timeval tv) {
  struct timeval tv1;
  int rval;
 
  gettimeofday(&tv1, NULL);
  rval = (tv1.tv_sec - tv.tv_sec) * 1000000;
  rval += (tv1.tv_usec - tv.tv_usec);
  
  return rval;
}

/* This function generates a random number for our private value, and computes 
   the public value as X = g^x mod n, sending the public value to the
   peer. This function will be used by both ends of the Diffie-Hellman
   keyexchange. We do not need to remember our public value */
static void dh_send_public(dh_obj_t *obj) {
  mpz_t public_mpz;
  unsigned char *public;
  control_msg_t *msg;
 
  random_uniform_2exp(obj->private_mpz, mpz_sizeinbase(obj->modulus_mpz,2)-1);
  if (mpz_cmp(obj->private_mpz,obj->modulus_mpz)>=0) {
#ifndef WIN32
    syslog(LOG_ERR,"DH Key exchange: Our private random number is larger"
	   " than the modulus. Oops (auto corrected)!\n");
#endif
    /* Corrent the problem -- right shift it one bit until it's smaller */
    while(mpz_cmp(obj->private_mpz,obj->modulus_mpz))
      mpz_fdiv_q_2exp(obj->private_mpz,obj->private_mpz,2);
  }

  
  mpz_init(public_mpz);
  mpz_powm(public_mpz, obj->generator_mpz, obj->private_mpz, 
	   obj->modulus_mpz);
  
  public = mpz_get_str(NULL, 16, public_mpz);
  
  msg = new_control_msg(DIFFIE_HELLMAN, DH_PUBLIC, obj->session_obj);
  msg_append_tlv(msg, DH_PUBLIC_VALUE, strlen(public), public);
  slan_memfree(public);
  mpz_clear(public_mpz);

  control_msg_send(msg);
  control_msg_destruct(msg);
}

/* This function waits for the peer to send it's public value. If received
   with non-zero length, we set the value in <obj>. Otherwise, the protocol
   aborts. This function is called by both ends of the exchange. */
static int dh_wait_public(dh_obj_t *obj) {
  struct timeval tv;
  control_msg_t *msg;
  int type, subtype;

  gettimeofday(&tv, NULL);
  do {
    msg = control_msg_receive(obj->session_obj, &type, &subtype, DH_TIME_MAX);
    if (type == DIFFIE_HELLMAN && subtype == DH_PUBLIC) {
      int public_length;
      unsigned char *public;

      msg_search_tlv(msg, DH_PUBLIC_VALUE, &public_length, &public);
      control_msg_destruct(msg);
      if (public_length != 0) {
	mpz_set_str(obj->public_mpz,public,16);
	slan_memfree(public);
	return 0;
      }
      else {
	u_int32_t value;
	value = DH_EBADPUBLIC;
	msg = new_control_msg(DIFFIE_HELLMAN, DH_ABORT, obj->session_obj);
	msg_append_tlv(msg, DH_ERROR, sizeof(value), 
		       (unsigned char *) &value);
	control_msg_send(msg);
	control_msg_destruct(msg);
	return DH_EBADPUBLIC;
      }
    }
    if (msg!=NULL) control_msg_destruct(msg);
  } while (dh_get_elapsed_time(tv) < DH_TIME_MAX);
  return DH_ETIMEOUT;
}

/* This function computes the shared key from the peer's public value and our
   private value. We convert the key to ascii hex and compute the hex string's 
   SHA digest. The digest is returned in hash_value. This will be used to
   check that the protocol completed correctly and the keys computed at each
   end match. THis function is called by both ends of the exchange. */
static void dh_compute_key(dh_obj_t *obj, unsigned char *hash_value) {
  int hash_document_length;
  unsigned char *hash_document;

  mpz_powm(obj->key_mpz, obj->public_mpz, obj->private_mpz, obj->modulus_mpz);

  /* May need some extra room for second level padding of hash*/
  hash_document_length = mpz_sizeinbase(obj->key_mpz,16) + 30;

  hash_document = slan_malloc(hash_document_length);
  memset(hash_document,0x3c,hash_document_length);
  mpz_get_str(hash_document,16,obj->key_mpz);

  sha_digest(hash_document, hash_document_length, hash_value);

  /* Do it again, protect against attacks on SHA and/or information leakage in 
     the hash about the actual key */
  memset(hash_document, 0x5c, hash_document_length);
  memcpy(hash_document + (hash_document_length/3), hash_value, 20);
  sha_digest(hash_document, hash_document_length, hash_value);

  slan_memfree(hash_document);
}

/* This function waits for the peer to send an acknowledgement, indicating
   that the hash value we sent for the key matches the one it computed,
   indicating that the key computed seperately at each end is the same, and
   the protocol can be completed. If we receive an DH_ABORT message, we assume 
   the key hash did not match, and the protocol did not complete 
   successfully.*/
static int dh_wait_keyack(dh_obj_t *obj) {
  struct timeval tv;
  control_msg_t *msg;
  int type, subtype;

  gettimeofday(&tv, NULL);
  do {
    msg = control_msg_receive(obj->session_obj, &type, &subtype, DH_TIME_MAX);
    if (type == DIFFIE_HELLMAN && subtype == DH_KEYACK) {
      control_msg_destruct(msg);
      return DH_FINISHED;
    }
    else if (type == DIFFIE_HELLMAN && subtype == DH_ABORT) {
      control_msg_destruct(msg);
      return DH_EBADHASH;
    }
    control_msg_destruct(msg);
  } while (dh_get_elapsed_time(tv) < DH_TIME_MAX);
  return DH_ETIMEOUT;
}

/* This function calls dh_compute_key() to compute the shared key between us
   and the peer, and sends the peer a portion of the SHA digest of the key
   (when converted to hex, see above). The hash is used to determine if the
   key computed at both ends is the same, without actually sending the key
   over the line (which would be stupid). Sending this hash value gives some
   information about the key to an evesdropper, so we only send 5 bytes of the 
   20 byte hash. Recontructing the key from the a portion of its SHA digest,
   should be computationally infeasible. */
static int dh_send_hash(dh_obj_t *obj) {
  control_msg_t *msg;
  unsigned char hash_value[20];

  dh_compute_key(obj, hash_value);

  msg = new_control_msg(DIFFIE_HELLMAN, DH_KEY, obj->session_obj);
  /* Only send part of the hash value -- we only need to verify that the
     calculated key on both ends is the same, we don't want to give out any
     more information about the key if we don't have to. */
  msg_append_tlv(msg, DH_KEYHASH, 5, hash_value);
  control_msg_send(msg);
  control_msg_destruct(msg);

  return dh_wait_keyack(obj);
}

static int dh_wait_keyhash(dh_obj_t *obj) {
  struct timeval tv;
  control_msg_t *msg;
  int type, subtype;
  int keyhash_length;
  unsigned char *keyhash;
  unsigned char hash_value[20];
  
  gettimeofday(&tv, NULL);
  do {
    msg = control_msg_receive(obj->session_obj, &type, &subtype, DH_TIME_MAX);
    if (type == DIFFIE_HELLMAN && subtype == DH_KEY) {
      dh_compute_key(obj, hash_value);
      msg_search_tlv(msg, DH_KEYHASH, &keyhash_length, &keyhash); 
      control_msg_destruct(msg);
      if (keyhash_length != 5 || memcmp(hash_value,keyhash,5)) {
	msg = new_control_msg(DIFFIE_HELLMAN, DH_ABORT, obj->session_obj);
	control_msg_send(msg);
	control_msg_destruct(msg);
	slan_memfree(keyhash);
	return DH_EBADHASH;
      } 
      else {
	msg = new_control_msg(DIFFIE_HELLMAN, DH_KEYACK, obj->session_obj);
	control_msg_send(msg);
	control_msg_destruct(msg);
	slan_memfree(keyhash);
	return DH_FINISHED;
      }
    }
    if (msg!=NULL) control_msg_destruct(msg);
  } while (dh_get_elapsed_time(tv) < DH_TIME_MAX);
  return DH_ETIMEOUT;
}

static void dh_extractkey(slan_session_t *session_obj, dh_obj_t *obj) {
  int byte;
  unsigned char *key_data;
  int key_length;

  /* This gets the key out of the mpz integer and makes it a long array of
     unsigned bytes. */
  key_length = mpz_sizeinbase(obj->key_mpz, CHAR_BIT);
  key_data = calloc(key_length+1, sizeof(unsigned char));
  for(byte=0;byte<key_length;byte++) {
    key_data[byte] = (unsigned char) (mpz_get_ui(obj->key_mpz) 
					      & UCHAR_MAX);
    mpz_fdiv_q_2exp(obj->key_mpz, obj->key_mpz, CHAR_BIT);
  }

  if (mpz_cmp_si(obj->key_mpz, UCHAR_MAX) > 0) {
#ifndef WIN32
    syslog(LOG_WARNING,"DH key warning: Key is longer than expected.\n");
    syslog(LOG_WARNING,"More than 8 bits discarded.\n");
#endif
  }
  
  /* Our keys, after this hashing compression, will be 160 bits long. We hope
     that this is conservative enough estimate of the entropy present in the
     DH key that was computed, which is about 1024 bits long in the usual 
     case and we hash half of it for each key. */
  session_obj->key = calloc(20, sizeof(unsigned char));
  session_obj->key_length = 20;
  sha_digest(key_data, key_length/2, session_obj->key);

  session_obj->mackey = calloc(20, sizeof(unsigned char));
  session_obj->mackey_length = 20;
  sha_digest(key_data + key_length/2, key_length/2, session_obj->mackey);

  /* To verify the key exchange, we use this hash of the entire key value
  computed, which is a triple SHA-1 with the last 10 bytes zero'd. The purpose
  of multiple hashing and zeroing the last 10 bytes is to generate something
  hard to guess, leaking as little key information as possible, but still long
  enough to be a strong enough check. This gives us 80 bits. */
  session_obj->keyhash = calloc(20, sizeof(unsigned char));
  session_obj->keyhash_length = 10;
  sha_digest(key_data, key_length,session_obj->keyhash);
  sha_digest(session_obj->keyhash, 20, session_obj->keyhash);
  sha_digest(session_obj->keyhash, 20, session_obj->keyhash);
  memset(session_obj->keyhash + 10, 0x0, 10);
  

  free(key_data);
}

/* This is the starting function for a server. It waits for a request to be
   sent and responds with values for g and n, as given in the configuration
   file */
int dh_listen(slan_session_t *session_obj) {
  struct timeval tv;
  control_msg_t *msg;
  dh_obj_t *obj;
  int type, subtype;
  int rval;

  gettimeofday(&tv, NULL);
  do {
    msg = control_msg_receive(session_obj, &type, &subtype, DH_TIME_MAX);
    if (type == DIFFIE_HELLMAN && subtype == DH_REQUEST) {
      unsigned char *modulus, *generator;
      control_msg_destruct(msg);
      msg = new_control_msg(DIFFIE_HELLMAN, DH_RESPONSE, session_obj);

      generator = slan_global_config->dh->gen;
      msg_append_tlv(msg, DH_GENERATOR, strlen(generator), generator);

      modulus = slan_global_config->dh->prime;
      msg_append_tlv(msg, DH_MODULUS, strlen(modulus), modulus);

      control_msg_send(msg);
      control_msg_destruct(msg);
      
      obj = new_dh_obj(session_obj);
      mpz_set(obj->generator_mpz, slan_global_config->dh->generator_mpz);
      mpz_set(obj->modulus_mpz, slan_global_config->dh->modulus_mpz);

      rval = dh_wait_public(obj);
      if (rval) {
	dh_obj_destruct(obj);
	return rval;
      } 
      dh_send_public(obj);
      rval = dh_wait_keyhash(obj);
      if (rval == DH_FINISHED) dh_extractkey(session_obj, obj);

      dh_obj_destruct(obj);
      return rval;
    }
    if (msg!=NULL) control_msg_destruct(msg);
  } while (dh_get_elapsed_time(tv) < DH_TIME_MAX);
#ifndef WIN32
  syslog(LOG_ERR,"DH listen timed out.\n");
#endif
  return DH_ETIMEOUT;
}

/* If the server sends us inappropriate values of g and n, we respond to the
   server aborting the protocol, indicating that g and n did not pass our
   tests. */
static void dh_send_badgn(dh_obj_t *obj) {
  u_int32_t value;
  control_msg_t *msg;

  msg = new_control_msg(DIFFIE_HELLMAN, DH_ABORT, obj->session_obj);
  value = DH_EBADGN;
  msg_append_tlv(msg, DH_ERROR, sizeof(value), (unsigned char *) &value);
  control_msg_send(msg);
  control_msg_destruct(msg);
}


/* This function tests values of g and n received, to ensure that the values
   are proper. In particular, n must be prime, and (n-1)/2 must also be
   prime. g must generate GF(n), we can determine easily because we know the
   factorization of (n-1). */
static int dh_test_gn(dh_obj_t *obj) {
  mpz_t temp, result;

  /* Check to see if this generator/prime pair is the same as the one in
     our configuration file. If so, don't bother with the expensive primality
     tests. */
  if (mpz_cmp(slan_global_config->dh->modulus_mpz, obj->modulus_mpz) == 0) {
    if (mpz_cmp(slan_global_config->dh->generator_mpz, obj->generator_mpz) 
	== 0) {
#ifndef WIN32
      syslog(LOG_INFO,"Server is using default generator/prime.\n");
#endif
      return 0;
    }
  }

#ifndef WIN32
  syslog(LOG_WARNING,"\nReceived unexpected g/n pair for DH-keyexchange.\n");
  syslog(LOG_WARNING,"Testing primality... this may take a moment.\n");
#endif
  fprintf(stderr,"\nReceived unexpected g/n pair for DH-keyexchange.\n");
  fprintf(stderr,"Testing primality... this may take a moment.\n");
  if (mpz_probab_prime_p(obj->modulus_mpz, 5)==0) return DH_EBADGN;
  
  mpz_init(temp);
  mpz_set(temp, obj->modulus_mpz);
  mpz_sub_ui(temp,temp,1);
  mpz_fdiv_q_2exp(temp,temp,1);
  
  if (mpz_probab_prime_p(temp, 5)==0) {
    mpz_clear(temp);
    return DH_EBADGN;
  }
  
  mpz_init(result);
  mpz_powm(result, obj->generator_mpz, temp, obj->modulus_mpz);
  if (mpz_cmp_ui(result,1) == 0) {
    mpz_clear(temp);
    mpz_clear(result);
    return DH_EBADGN;
  }
  
  mpz_set_ui(temp,2);
  mpz_powm(result, obj->generator_mpz, temp, obj->modulus_mpz);
  if (mpz_cmp_ui(result,1) == 0) {
    mpz_clear(temp);
    mpz_clear(result);
    return DH_EBADGN;
  }
  
  /* Generator and Modulus have passed requisite tests */
  mpz_clear(temp);
  mpz_clear(result);
  return 0;
}

/* This function waits for a response to a DH_REQUEST message. THe response
   should be a DH_RESPONSE, containing values for g and n, or a DH_DENY,
   indicating that the server does not which to use Diffie-Hellman key
   exchange (or that it is not a server). DH_DENY causes the protocol to
   abort. */
static int dh_wait_gn(dh_obj_t *obj) {
  struct timeval tv;
  control_msg_t *msg;
  int type, subtype;

  gettimeofday(&tv, NULL);
  do {
    msg = control_msg_receive(obj->session_obj, &type, &subtype, DH_TIME_MAX);
    if (msg!=NULL) {
      if (type == DIFFIE_HELLMAN && subtype == DH_RESPONSE) {
	int g_length, n_length;
	unsigned char *g, *n;
	int rval;

	msg_search_tlv(msg, DH_GENERATOR, &g_length, &g);
	msg_search_tlv(msg, DH_MODULUS, &n_length, &n);
	control_msg_destruct(msg);
	if (g_length == 0 || n_length == 0) {
	  dh_send_badgn(obj);
	  if (g) slan_memfree(g);
	  if (n) slan_memfree(n);
	  return DH_EBADGN;
	}
	mpz_set_str(obj->modulus_mpz, n, 0);
	mpz_set_str(obj->generator_mpz, g, 0);
	slan_memfree(g); slan_memfree(n);

	/* Test to make sure the values for g and n are acceptable for Diffie
	   Hellman keyexchange. If not, we send a message to the server saying 
	   that we refuse the values of g and n, and abort the protocol. */
	rval = dh_test_gn(obj);
	if (rval == DH_EBADGN) {
	  dh_send_badgn(obj);
	  return DH_EBADGN;
	}

	/* The values for g and n are okay, send the server our public 
	   value. */
	dh_send_public(obj);
	rval =  dh_wait_public(obj);
	if (rval) return rval;
	return dh_send_hash(obj);
      }
      /* We may also receive a DENY message from the server, indicating that
	 either it is unwilling or cannot (resource contraints suppose) to
	 perform the keyexchange. This aborts the protocol. */
      else if (type == DIFFIE_HELLMAN && subtype == DH_DENY) {
	control_msg_destruct(msg);
	return DH_EDENY;
      }

      /* Ignore all other messages */
      control_msg_destruct(msg);
    }
  } while (dh_get_elapsed_time(tv) < DH_TIME_MAX);
  /* If the server cannot respond appropriately in a timely fashion, we abort
     the protocol. This may indicate either an overloaded server, network
     problems, or attack. Any of these means we cannot or should not continue
     the protocol. */
  return DH_ETIMEOUT;
}

/* This is the starting function for a client who initiates a diffie-hellman
   key exchange. We start by sending a request to the server, which should
   respond with values for g and n. */
int dh_initiate(slan_session_t *session_obj) {
  control_msg_t *msg;
  dh_obj_t *obj;
  int return_value;

  msg = new_control_msg(DIFFIE_HELLMAN, DH_REQUEST, session_obj);
  control_msg_send(msg);
  control_msg_destruct(msg);

  obj = new_dh_obj(session_obj);

  return_value = dh_wait_gn(obj);
  if (return_value != DH_FINISHED) {
    /* TODO: Place Error handling here */
    fprintf(stderr,"Diffie-Hellman key exchange failure: %s\n",
	    dh_error[return_value]);
#ifndef WIN32
    syslog(LOG_ERR,"Diffie-Hellman key exchange failure (%s)\n",
	   dh_error[return_value]);
#endif
    dh_obj_destruct(obj);
    return return_value;
  }
  else {
    dh_extractkey(session_obj, obj);
  }

  dh_obj_destruct(obj);
  return 0;
}

