/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage4/control/ip_conf.c,v 1.6 2001/05/08 15:19:44 koni Exp $ 
  $Author: koni $ $Date: 2001/05/08 15:19:44 $ $Name:  $
  $Revision: 1.6 $

*/
   
#ifndef WIN32
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#else
#include "slan_windows.h"
#endif

#include "slan_types.h"
#include "slan_alloc.h"

#include "ip_conf.h"
#include "slan_session.h"

#include "control_msg.h"
#include "control_tcp.h"


/* Number of microseconds before ip configuration fails -- 10 seconds */
#define IPCONF_TIME_MAX (10000000)

enum { SESSION_ID = 0, IP_ADDRESS, SUBNET_MASK, SUBNET_BCAST, ROUTER, 
       NAMESERVER1, NAMESERVER2, N_IPCONF_ITEMS };

enum { IPCONF_REQ = 0, IPCONF_INF, IPCONF_FAIL, N_IPCONF_MSG };

/* This (stupid) function wraps calls to free, so that we can call free
   when we expect to free memory that should have been allocated by another
   function, but, may in fact have not been allocated (and p == NULL) in
   exceptional or unexpected cases. */
static void safe_free(void *p) {
  if (p!=NULL) slan_memfree(p);
}

static int ipconf_get_elapsed_time(struct timeval tv) {
  struct timeval tv1;
  int rval;
 
  gettimeofday(&tv1, NULL);
  rval = (tv1.tv_sec - tv.tv_sec) * 1000000;
  rval += (tv1.tv_usec - tv.tv_usec);
  
  return rval;
}

/* This function (called by the client) sends an ip configuration request 
   and waits/interprets the response. All fields are expected to length 4 --
   if any of them are not (except nameserver2), we return nonzero indicating
   failure. */
int ipconf_request(ip_info_t *ip_conf, slan_session_t *session_obj) {
  struct timeval tv;
  int type, subtype;
  control_msg_t *msg;
  unsigned char *addr;
  int length;

  msg = new_control_msg(IPCONF, IPCONF_REQ, session_obj);
  control_msg_send(msg);
  control_msg_destruct(msg);

  gettimeofday(&tv, NULL);
  do {
    msg = control_msg_receive(session_obj, &type, &subtype, IPCONF_TIME_MAX);
    if (msg != NULL) {
      if (type == IPCONF && subtype == IPCONF_FAIL) {
	control_msg_destruct(msg);
	return 1;
      }
      if (type == IPCONF && subtype == IPCONF_INF) {

	msg_search_tlv(msg, SESSION_ID, &length, &addr);
	if (length==4)
	  memcpy(&session_obj->id,addr,4);
	safe_free(addr);

	msg_search_tlv(msg, IP_ADDRESS, &length, &addr);
	if (length==4)
	  memcpy(&ip_conf->ip_address,addr,4);
	safe_free(addr);
	
	msg_search_tlv(msg, SUBNET_MASK, &length, &addr);
	if (length==4)
	  memcpy(&ip_conf->subnet_mask,addr,4);
	safe_free(addr);
	  
	msg_search_tlv(msg, SUBNET_BCAST, &length, &addr);
	if (length==4)
	  memcpy(&ip_conf->subnet_bcast,addr,4);
	safe_free(addr);

	msg_search_tlv(msg, ROUTER, &length, &addr);
	if (length==4)
	  memcpy(&ip_conf->router,addr,4);
	safe_free(addr);

	msg_search_tlv(msg, NAMESERVER1, &length, &addr);
	if (length==4)
	  memcpy(&ip_conf->nameserver_1,addr,4);
	safe_free(addr);

	msg_search_tlv(msg, NAMESERVER2, &length, &addr);
	if (length==4)
	  memcpy(&ip_conf->nameserver_2,addr,4);
	safe_free(addr);

	control_msg_destruct(msg);
	return 0;
      }	
      control_msg_destruct(msg);
    }
  } while (ipconf_get_elapsed_time(tv) < IPCONF_TIME_MAX);

  return 1;
}

/* This function, called by the server, waits for a request from the client
   and responds with the ip address and associated information, passed in 
   by the ip_info_t structure. */
int ipconf_req_wait(ip_info_t *ip_conf, slan_session_t *session_obj) {
  struct timeval tv;
  int type, subtype;
  control_msg_t *msg;

  gettimeofday(&tv, NULL);
  do {
    msg = control_msg_receive(session_obj, &type, &subtype, IPCONF_TIME_MAX);
    /* We are not expecting any data in the message, so we can free it 
       right away if we get one */
    if (msg!=NULL) control_msg_destruct(msg);

    if (type == IPCONF && subtype == IPCONF_REQ) {
      if (ip_conf == NULL) {
	/* Failed to get ip configuration information */
	msg = new_control_msg(IPCONF, IPCONF_FAIL, session_obj);
	control_msg_send(msg);
	control_msg_destruct(msg);
	return 0;
      }

      msg = new_control_msg(IPCONF, IPCONF_INF, session_obj);

      msg_append_tlv(msg, SESSION_ID, 4, (unsigned char *)
		     &session_obj->id);      
      msg_append_tlv(msg, IP_ADDRESS, 4, (unsigned char *) 
		     &ip_conf->ip_address);
      msg_append_tlv(msg, SUBNET_MASK, 4, (unsigned char *) 
		     &ip_conf->subnet_mask);
      msg_append_tlv(msg, SUBNET_BCAST, 4, (unsigned char *)
		     &ip_conf->subnet_bcast);
      msg_append_tlv(msg, ROUTER, 4, (unsigned char *) &ip_conf->router);
      msg_append_tlv(msg, NAMESERVER1, 4, (unsigned char *)
		     &ip_conf->nameserver_1);
      /* Second nameserver is an optional configuration parameter, it may
	 not be present. */
      if (ip_conf->nameserver_2 != 0) 
	msg_append_tlv(msg, NAMESERVER2, 4, (unsigned char *)
		       &ip_conf->nameserver_2);
      control_msg_send(msg);
      control_msg_destruct(msg);
      return 0;
    }
  } while (ipconf_get_elapsed_time(tv) < IPCONF_TIME_MAX);
  
  return 1;
}

