/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage4/control/server_pamauth.c,v 1.12 2001/06/25 02:12:21 koni Exp $ 
  $Author: koni $ $Date: 2001/06/25 02:12:21 $ $Name:  $
  $Revision: 1.12 $

*/
   
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <security/pam_appl.h>

#include "slan_alloc.h"

#include "slan_session.h"
#include "control_tcp.h"
#include "control_msg.h"
#include "client_pamauth.h"

/* The documentation on how this function is supposed to work with PAM is
   quite unclear -- I certainly hope PAM is going to free the memory
   allocated. 

   PAM calls this when it wants to ask the client questions. We gobble up
   these questions and send them to the client. The client will answer those
   messages which were prompts and do whatever with the others. */
static int slan_pamconv(int num_msg, const struct pam_message **pam_msg,
			 struct pam_response **resp, void *sesobj) {
  slan_session_t *session_obj;
  struct pam_response *pam_resp;
  int i;
  control_msg_t *msg;
  u_int32_t type, sub_type;
  int length;
  unsigned char *response;
  struct timeval tv;

  /* We tell the pam library to pass us the session obj when this conversation 
     function is called. Thus, we avoid need for global state. */
  session_obj = (slan_session_t *) sesobj;

  msg = new_control_msg(CLIENT_AUTH, PAM_SENDCLIENT, session_obj);

  type = htonl(num_msg);
  msg_append_tlv(msg, NUM_MSG, sizeof(u_int32_t), (unsigned char *) &type);

  for(i=0;i<num_msg;i++) {
    switch(pam_msg[i]->msg_style) {
    case PAM_PROMPT_ECHO_OFF:
      type = htonl(PROMPT_NOECHO);
      msg_append_tlv(msg, PAM_MSG + i, sizeof(u_int32_t), (unsigned char *)
		     &type);
      msg_append_tlv(msg, MSG_DATA + i, strlen(pam_msg[i]->msg), 
		     (unsigned char *) pam_msg[i]->msg);
      break;
    case PAM_PROMPT_ECHO_ON:
      type = htonl(PROMPT_ECHO);
      msg_append_tlv(msg, PAM_MSG + i, sizeof(u_int32_t), (unsigned char *)
		     &type);
      msg_append_tlv(msg, MSG_DATA + i, strlen(pam_msg[i]->msg), 
		     (unsigned char *) pam_msg[i]->msg);
      break;
    case PAM_ERROR_MSG:
      type = htonl(WARNING);
      msg_append_tlv(msg, PAM_MSG + i, sizeof(u_int32_t), (unsigned char *)
		     &type);
      msg_append_tlv(msg, MSG_DATA + i, strlen(pam_msg[i]->msg), 
		     (unsigned char *) pam_msg[i]->msg);
      break;
    case PAM_TEXT_INFO:
      type = htonl(INFORMATION);
      msg_append_tlv(msg, PAM_MSG + i, sizeof(u_int32_t), (unsigned char *)
		     &type);
      msg_append_tlv(msg, MSG_DATA + i, strlen(pam_msg[i]->msg), 
		     (unsigned char *) pam_msg[i]->msg);
      break;
    default:
      syslog(LOG_ERR,"Client AUTH error: Unknown pam message %d\n", 
	     pam_msg[i]->msg_style);
      type = htonl(SLANPAM_UNKNOWN);
      msg_append_tlv(msg, PAM_MSG + i, sizeof(u_int32_t), (unsigned char *)
		     &type);
      msg_append_tlv(msg, MSG_DATA + i, strlen(pam_msg[i]->msg), 
		     (unsigned char *) pam_msg[i]->msg);
      break;
    }
  }
  control_msg_send(msg);
  control_msg_destruct(msg);

  gettimeofday(&tv, NULL);
  do {
    msg = control_msg_receive(session_obj, &type, &sub_type, 
			      CLIENTAUTH_TIMEOUT);
    if (type == CLIENT_AUTH && sub_type == PAM_CLIENTRESP) {
      pam_resp = slan_calloc(num_msg, sizeof(struct pam_response));
      for(i=0;i<num_msg;i++) {
	if (pam_msg[i]->msg_style == PAM_PROMPT_ECHO_OFF || 
	    pam_msg[i]->msg_style == PAM_PROMPT_ECHO_ON) {
	  response = NULL;
	  msg_search_tlv(msg, PAM_RESPONSE + i, &length, &response);
	  if (length > 0) {
	    /* May not be a string, byte copy instead of string copy */
	    pam_resp[i].resp = response;
	    //	    memcpy(pam_resp[i].resp, response, length);
	    
	    pam_resp[i].resp_retcode = 0;
	  } else {
	    if (response) slan_memfree(response);
	    pam_resp[i].resp = NULL;
	    pam_resp[i].resp_retcode = 0;
	  }
	} else {
	  pam_resp[i].resp = NULL;
	  pam_resp[i].resp_retcode = 0;
	}
      }
      control_msg_destruct(msg);
      *resp = pam_resp;
      return PAM_SUCCESS;
    }	
    control_msg_destruct(msg);
  } while (control_get_elapsed_time(tv) < CLIENT_AUTH_TIMEMAX);
  return PAM_CONV_ERR;
}

static int server_pam_authenticate(slan_session_t *session_obj) {
  pam_handle_t *pamh;
  struct pam_conv slan_conv;
  control_msg_t *msg;
  int retval;

  slan_conv.conv = slan_pamconv;
  slan_conv.appdata_ptr = session_obj; 

  retval = pam_start("slan", NULL, &slan_conv, &pamh);

  if (retval != PAM_SUCCESS) {
    syslog(LOG_ERR,"Client Authentication Failure: pam_start(): %s\n",
	    pam_strerror(pamh, retval));
    msg = new_control_msg(CLIENT_AUTH, AUTH_REJECT, session_obj);
    control_msg_send(msg);
    control_msg_destruct(msg);
    pam_end(pamh, PAM_SUCCESS);
    return 1;
  }

  retval = pam_authenticate(pamh, PAM_SILENT);
  if (retval != PAM_SUCCESS) {
    syslog(LOG_ERR,"Client Authentication Failure: pam_authenticate(): %s\n",
	   pam_strerror(pamh, retval));
    msg = new_control_msg(CLIENT_AUTH, AUTH_REJECT, session_obj);
    control_msg_send(msg);
    control_msg_destruct(msg);
    pam_end(pamh, PAM_SUCCESS);
    return 1;
  }

  pam_end(pamh, PAM_SUCCESS);
  /* Later, we may want to do something a bit more fancy, like check the
     user's credentials and/or open a session for the user and use that for
     accounting, etc. */ 
  msg = new_control_msg(CLIENT_AUTH, AUTH_OKAY, session_obj);
  control_msg_send(msg);
  control_msg_destruct(msg);

  return 0;
}

/* This function is called when server wishes to authenticate the client. It
   tells the client that authentication is going to occur -- if the client
   responds that it is ready, it proceeds with PAM authentication. */
int server_clientauth_initiate(slan_session_t *session_obj) {
  control_msg_t *msg;
  struct timeval tv;
  int type, sub_type;

  /* Tell client we are going to authenticate it */
  msg = new_control_msg(CLIENT_AUTH, AUTH_REQUEST, session_obj);
  control_msg_send(msg);
  control_msg_destruct(msg);
  
  /* Wait for the client to say its ready, expecting client authentication to
     proceed. */
  gettimeofday(&tv, NULL);
  do {
    msg = control_msg_receive(session_obj, &type, &sub_type, 
			      CLIENTAUTH_TIMEOUT);
    if (type == CLIENT_AUTH && sub_type == AUTH_READY) {
      control_msg_destruct(msg);
      return server_pam_authenticate(session_obj);
    }
    control_msg_destruct(msg);
  } while(control_get_elapsed_time(tv) < CLIENT_AUTH_TIMEMAX);
  return 1;
}
    

