/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage4/platform/linux/slan_client.c,v 1.15 2001/07/13 03:14:52 koni Exp $ 
  $Author: koni $ $Date: 2001/07/13 03:14:52 $ $Name:  $
  $Revision: 1.15 $

*/
   
/* The Lightlink Secure LAN project

   KONI - December 2000
   slan_client.c

   Linux client program
*/

/* Standard C library includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <syslog.h>
#include <errno.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/* to secure memory (lock from paging to disk) */
#include <sys/mman.h>

/* Network stuff */
#include <sys/socket.h>
#include <sys/poll.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <net/route.h>

/* GNU multiprecision math library */
#ifdef HAVE_GMP_H
#include <gmp.h>
#else
#include "slan_gmp3.h"
#endif

/* session state type definitions and constants */
#include "slan_session.h"
#include "slan_version.h"
#include "slan_config.h"

/* memory wrapper */
#include "slan_alloc.h"

/* Cryptography */
#include "blowfish.h"
#include "sym_cipher.h"
#include "sha.h"
#include "mac.h"
#include "hmac_sha1.h"

/* Setup and Control Protocol components */
#include "dh_keyexchange.h"
#include "rsaauth.h"
#include "random_interface.h"
#include "client_auth.h"
#include "ip_conf.h"
#include "control_msg.h"
#include "control_tcp.h"
#include "client_pamauth.h"

/* Transport protocol stack */
#include "proto_stack.h"
#include "stack_register.h"

/* This variable is global to the entire slan program */
slan_conf_t *slan_global_config;

/* These file-scope variables are used to indicate shutdown status (requested
   by signal from operating system) and whether or not to configure the slan0
   interface and fork, or just connect and drop (development only) */
static volatile int shutdown_now = 0;
static int connect_only = 0;

/* This is set if the client terminates because the server hung up or an
   exceptional circumstance occurs. This prevents the shutdown process from
   restoring default routing for the real network. Otherwise, the user may
   continue using the network under false assumption that the security
   measures in SLAN are still in effect. */
static int dont_restore = 0;

/* Main state information for the SLAN session is stored in this struct */
static slan_session_t *session_obj;

#define PID_FILENAME "/var/tmp/slan.pid"
#define PIPE_FILENAME "/var/tmp/.slan_statuspipe"

#define NOBODY_UID (99)
#define NOBODY_GID (99)

static void client_setup_transport_layer(slan_session_t *session_obj,
					 char *server_address) {
  proto_stack_t *last, *current;
  int header_save;

  header_save = 0;
  /* A stack terminator at the top -- this is simply a dummy protocol
     layer that drops any packet passed to it. It can be used to catch
     bad or dropped packets and debug them. Otherwise, it isn't necessary and
     doesn't do much. */
  current = last = slan_malloc(sizeof(proto_stack_t));
  terminate_register(current);
  header_save += current->header_reserve;
  
  /* Packets received from the slan device sent down the stack from the top */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  slan_link_register(current, 0, session_obj);
  header_save += current->header_reserve;
  session_obj->transport_top = current;
  current->prev = last;
  last = current;

  /* Optional fragmenting protocol layer -- currently not implemented */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  fragment_register(current);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* All packets being sent through the slan tunnel get tagged with sequence
     numbers before adding a MAC and encrypting. This used to limit the
     possibilities for message replay attacks. */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  sequence_register(current);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* After adding the sequence number, we add a Message Authentication Code
     (MAC) to allow the receiver to verify the integrity of the packet and the 
     sender of the data */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  message_sign_register(current, session_obj->mac);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* Lastly, we enrypt the packet for privacy during transmission. Currently,
     I am expecting the symmetric cipher used to be a configurable option, so
     the cipher used depends on the field in the session_obj */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  symmetric_cipher_register(current, session_obj->cipher);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* Session ID numbers are prepended to identify the session the message
     belongs to */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  session_id_register(current);
  header_save += current->header_reserve;
  *((u_int32_t *) current->data) = session_obj->id;
  current->prev = last;
  last = current;

  /* Transmit the secured packet via the UDP channel */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  client_udplink_register(current, server_address, 
			  slan_global_config->transport_port, session_obj);
  header_save += current->header_reserve;
  session_obj->transport_bottom = current;
  current->prev = last;
  last = current;

  /* Terminate the bottom of the stack */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  terminate_register(current);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  session_obj->header_save = header_save;
}

static void forget_keys(slan_session_t *session_obj) {
  int i;

  /* Forget keys. NULL pointers may exist if shutdown happens before the 
     session is configured fully */
  if (session_obj->mackey!=NULL) {
    for(i=0;i<session_obj->mackey_length;i++)
      session_obj->mackey[i] = 0x0;
    free(session_obj->mackey);
  }
  session_obj->mackey = NULL;

  if (session_obj->key!=NULL) {
    for(i=0;i<session_obj->key_length;i++)
      session_obj->key[i] = 0x0;
    slan_memfree(session_obj->key);
  }
  session_obj->key = NULL;

  if (session_obj->keyhash!=NULL) {
    for(i=0;i<session_obj->keyhash_length;i++)
      session_obj->keyhash[i] = 0x0;
    slan_memfree(session_obj->keyhash);
  }
  session_obj->keyhash = NULL;

}

static void client_shutdown(slan_session_t *session_obj) {
  
  syslog(LOG_INFO,"Connection closing...\n");

  session_obj->cipher->close(session_obj->cipher);
  session_obj->mac->close(session_obj->mac);
  forget_keys(session_obj);

  /* Transport layer destruction should go here */

  /* Closing the control channel (TCP) socket alerts the server to the fact
     that we are shuting down. */
  close(session_obj->control_fd);

  /* This should cause the process to exit */
  pthread_exit(NULL);
}

static void change_key(slan_session_t *session_obj) {

  forget_keys(session_obj);
  if (dh_initiate(session_obj)) {
    syslog(LOG_ERR,"Change session key: DH exchange failed!\n");
    syslog(LOG_ERR,"The key was not changed.\n");
  } else {
    syslog(LOG_INFO,"Setting new MAC and cipher keys....\n"); fflush(stdout);
    session_obj->cipher->set_key(session_obj->cipher, session_obj->key,
				 session_obj->key_length); 
    
    session_obj->mac->set_key(session_obj->mac, session_obj->mackey,
			      session_obj->mackey_length); 
    /* Reset sequence numbers: Maybe this should be randomized a bit */
    session_obj->sequence_down = 0;
    session_obj->sequence_up = 0;
    
    session_obj->key_timestamp = time(NULL);
    /* Unblock the transport layer. */
  }
  session_obj->key_change = 0;
}

/* This is called by the signal handler */
static void interrupt_shutdown(int signal) {
  shutdown_now = 1;
}

static void client_unexpected_shutdown(slan_session_t *session_obj) {

  syslog(LOG_EMERG,"\n\nThe connection to the server has unexpectedly "
	  "closed.\n");
  syslog(LOG_EMERG,"The session will be terminated (sorry!).\n\n");

  /* Don't restore default routes -- force user to know (network will stop
     working) that SLAN has terminated without user's request */
  dont_restore = 1;

  /* !! This should restart the connection automatically, if possible, instead 
        of full shutdown. !! */
  client_shutdown(session_obj);
}


static void *control_loop(void *arg) {
  slan_session_t *session_obj;
  struct pollfd ufds;
  control_msg_t *msg, *new_msg;
  u_int32_t type, subtype;

  session_obj = arg;
  for(;;) {
    ufds.fd = session_obj->control_fd;
    ufds.events = ( POLLIN | POLLPRI | POLLERR | POLLHUP );
    ufds.revents = 0;

    /* Give up after 2 seconds, check to see if we should do anything
       and then go back to polling. */
    poll(&ufds, 1, 2000);
    pthread_mutex_lock(session_obj->lock);
    if (shutdown_now) {
      syslog(LOG_INFO,"Interrupt Received, shutdown started.\n");
      client_shutdown(session_obj);
    }

    if (ufds.revents == POLLHUP) {
      /* If the server hangs up, we shutdown immediately. */
      client_unexpected_shutdown(session_obj);
    }
    if (ufds.revents == POLLPRI) {
      /* Our protocol doesn't use out of band data. So we just ignore it */
      syslog(LOG_ERR,"Out of band data waiting... why is that?\n");
      continue;
    }
    if (ufds.revents == POLLERR) {
      /* I don't know what might cause this to occur, so for now, we'll
	 shutdown the connection if we have an error */ 
      syslog(LOG_ERR,"Error condition on control channel.");
      client_shutdown(session_obj);
    }
    if (ufds.revents == POLLIN) {
      msg = control_msg_receive(session_obj, &type, &subtype, 0);
      if (msg == NULL) {
	pthread_mutex_unlock(session_obj->lock);
	continue;
      }

      switch(type) {
      case KEYCHANGE:
	if (subtype == KEYCHANGE_READY) {
	  change_key(session_obj);
	  control_msg_destruct(msg);
	} else {
	  session_obj->key_change = 1;
	  new_msg = new_control_msg(KEYCHANGE, KEYCHANGE_READY, session_obj);
	  control_msg_send(new_msg);
	  pthread_mutex_unlock(session_obj->lock);
	  control_msg_destruct(new_msg);
	  control_msg_destruct(msg);
	  usleep(1000);
	  pthread_mutex_lock(session_obj->lock);
	  change_key(session_obj);
	  
	}
	pthread_mutex_unlock(session_obj->lock);
	continue;
	break;
      case SHUTDOWN:
	/* This should not return */
	client_shutdown(session_obj);
	break;
      case KEEPALIVE:
	if (subtype != KEEPALIVE_PING) {
	  /* This should not happen. */
	  syslog(LOG_ERR,"Not expecting a keepalive response here.\n");
	} else {
	  new_msg = new_control_msg(KEEPALIVE, KEEPALIVE_RESPONSE, 
				    session_obj);
	  control_msg_send(new_msg);
	  control_msg_destruct(new_msg);
	}
	break;
      default:
	syslog(LOG_CRIT,"Unhandled control message type: %d\n",type);
	break;
      }
      control_msg_destruct(msg);
    }
    
    /* Check the session key's age. Change it if it should be expired. This is
       the client's responsibility to enforce. */
    if ((time(NULL) - session_obj->key_timestamp) > KEY_LIFETIME &&
	session_obj->key_change == 0) {
      session_obj->key_change = 1;
      new_msg = new_control_msg(KEYCHANGE, KEYCHANGE_REQUEST, session_obj);
      control_msg_send(new_msg);
      control_msg_destruct(new_msg);
    }
    pthread_mutex_unlock(session_obj->lock);
  }
}

/* This function continuouly reads either UDP socket for incoming traffic or
   the SLAN socket for outgoing traffic, using the builtin transport layer
   functions. They are non-blocking, using select to test for data before
   reading.  

   TODO: Make this a multithreaded operation and allow blocking calls?
   Performance at the client end seems to be fine with the current
   implementation. */
/* Incremental sleeping step size. The longer we have nothing to do the more
   we wait before looking for packets to process */
#define SLEEP_STEP (1000)
#define SLEEP_MAX  (1000)
/* Maximum number of packets to process in one direction before switching to
   the other. Blocking will cause immediate switch. */
#define PACKETS_PER_ATTEMPT (10)
static void *transport_loop(void *arg) {
  slan_session_t *session_obj;
  unsigned char buf[10240];
  unsigned char *p;
  slan_message_t slan_message;
  int packet;
  int slan_socket, udp_socket, max_fd;
  fd_set read_fds;

  session_obj = arg;

  slan_message.buffer = buf;
  slan_message.maxspace = 4096;

  p = buf + session_obj->header_save;
  
  slan_message.header = slan_message.data = p;
  slan_message.trailer = p;
  slan_message.length = 0;

  udp_socket = session_obj->udp_socket;
  slan_socket = session_obj->slan_socket;
  max_fd = ((udp_socket > slan_socket)?udp_socket:slan_socket) + 1;

  FD_ZERO(&read_fds);
  
  for(;;) {
    FD_SET(udp_socket, &read_fds);
    FD_SET(slan_socket, &read_fds);
    
    select(max_fd, &read_fds, NULL, NULL, NULL);
    
    pthread_mutex_lock(session_obj->lock);

    if (FD_ISSET(slan_socket, &read_fds) && !session_obj->key_change) {
      session_obj->transport_top->down(session_obj->transport_top, 
				       &slan_message);
      slan_message.header = slan_message.data = p;
      slan_message.trailer = p;
      slan_message.length = 0;
    }
    
    if (FD_ISSET(udp_socket, &read_fds)) {
      session_obj->transport_bottom->up(session_obj->transport_bottom, 
					&slan_message);
      slan_message.header = slan_message.data =p;
      slan_message.trailer = p;
      slan_message.length = 0;
    }
    pthread_mutex_unlock(session_obj->lock);
  }   
}


/* Temporary -- direct IP configuration code */

/* Get ready to reach around your ass to grab the other elbow.... */
static int  linux_configure_network(ip_info_t *ip_conf, 
				    unsigned int server_route,
				    unsigned int server_addr,
				    int sock)  { 
  struct ifreq ifr;
  struct sockaddr_in sa;
  struct rtentry rt;
  struct in_addr ns;
  FILE *f;

  strcpy(ifr.ifr_name,"slan0");

  sa.sin_family = AF_INET;
  memcpy(&sa.sin_addr, &ip_conf->ip_address, 4);
  memcpy(&ifr.ifr_addr, &sa, sizeof(struct sockaddr));
  if (ioctl(sock, SIOCSIFADDR, &ifr) < 0) {
    fprintf(stderr,"Could not set ipaddress for interface.\n");
    perror("Error code:");
    return 1;
  } 

  memcpy(&sa.sin_addr, &ip_conf->subnet_mask, 4);
  memcpy(&ifr.ifr_netmask, &sa, sizeof(struct sockaddr));

  if (ioctl(sock, SIOCSIFNETMASK, &ifr) < 0) {
    fprintf(stderr,"Could not set subnet mask for interface.\n");
    perror("Error code:");
    return 1;
  } 

  memcpy(&sa.sin_addr, &ip_conf->subnet_bcast, 4);
  memcpy(&ifr.ifr_broadaddr, &sa, sizeof(struct sockaddr));
  if (ioctl(sock, SIOCSIFBRDADDR, &ifr) < 0) {
    fprintf(stderr,"Could not set broadcast for interface.\n");
    perror("Error code:");
    return 1;
  } 

  /* Start the interface going like a normal ethernet except we don't do arp */
  ifr.ifr_flags = IFF_UP | IFF_RUNNING | IFF_NOARP | IFF_BROADCAST;
  if (ioctl(sock, SIOCSIFFLAGS, &ifr) < 0) {
    fprintf(stderr,"Failed setting interface flags.\n");
    perror("Error code:");
    return 1;
  } 
  
  /* Now do routing */
  
  /* Clear the route entry structure */
  memset(&rt, 0x0, sizeof(struct rtentry));
  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = htonl(INADDR_ANY);
  
  memcpy(&rt.rt_dst, &sa, sizeof(struct sockaddr_in));
  memcpy(&rt.rt_genmask, &sa, sizeof(struct sockaddr_in));
  
  rt.rt_flags = RTF_UP | RTF_GATEWAY;
  while(ioctl(sock, SIOCDELRT, &rt) == 0) {
    syslog(LOG_INFO,"Deleted default route from routing table.\n");
  }

  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = server_addr;
  memcpy(&rt.rt_dst, &sa, sizeof(struct sockaddr_in));

  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = 0xFFFFFFFF;
  memcpy(&rt.rt_genmask, &sa, sizeof(struct sockaddr_in));

  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = server_route;
  memcpy(&rt.rt_gateway, &sa, sizeof(struct sockaddr_in));

  rt.rt_flags = RTF_HOST | RTF_GATEWAY | RTF_UP;
  if (ioctl(sock, SIOCADDRT, &rt) < 0) {
    perror("Could not set routing information for slan server route.\n");
    return 1;
  }

  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = 0x0;
  memcpy(&rt.rt_dst, &sa, sizeof(struct sockaddr_in));

  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = 0x0;
  memcpy(&rt.rt_genmask, &sa, sizeof(struct sockaddr_in));

  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = ip_conf->router;
  memcpy(&rt.rt_gateway, &sa, sizeof(struct sockaddr_in));

  rt.rt_flags = RTF_GATEWAY | RTF_UP;
  if (ioctl(sock, SIOCADDRT, &rt) < 0) {
    perror("Could set default gateway in routing table.\n");
    return 1;
  }
  close(sock);
  
  /* Set nameserver configuration */
  f = fopen("/etc/resolv.conf","w");
  if (f == NULL) return 1;

  ns.s_addr = ip_conf->nameserver_1;
  fprintf(f,"nameserver %s\n",inet_ntoa(ns));
  if (ip_conf->nameserver_2) {
    ns.s_addr = ip_conf->nameserver_2;
    fprintf(f,"nameserver %s\n",inet_ntoa(ns));
  }
  fclose(f);

  return 0;
}

static void client_restore_routes(unsigned int server_route,
				  unsigned int server_addr) {
  int sock;
  struct sockaddr_in sa;
  struct rtentry rt;
  struct ifreq ifr;

  sock = socket(AF_INET, SOCK_DGRAM, 0);

  if (sock < 0) {
    syslog(LOG_ERR,"Cannot open socket for routing table configuration (%m)");
    return;
  }

  /* Clear the route entry structure */
  memset(&rt, 0x0, sizeof(struct rtentry));
  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = htonl(INADDR_ANY);
  
  memcpy(&rt.rt_dst, &sa, sizeof(struct sockaddr_in));
  memcpy(&rt.rt_genmask, &sa, sizeof(struct sockaddr_in));
  
  rt.rt_flags = RTF_UP | RTF_GATEWAY;
  if (ioctl(sock, SIOCDELRT, &rt)<0){
    syslog(LOG_ERR,"Failed deleting default route (%s)\n", strerror(errno));
  }

  if (!dont_restore) {
    sa.sin_addr.s_addr = server_route;
    memcpy(&rt.rt_gateway, &sa, sizeof(struct sockaddr_in));
    if (ioctl(sock, SIOCADDRT, &rt)<0) {
      syslog(LOG_ERR,"Failed to restore default routing.\n");
    }
  }

  sa.sin_addr.s_addr = server_addr;
  memcpy(&rt.rt_dst, &sa, sizeof(struct sockaddr_in));

  sa.sin_addr.s_addr = 0xFFFFFFFF;
  memcpy(&rt.rt_genmask, &sa, sizeof(struct sockaddr_in));

  rt.rt_flags |= RTF_HOST;
  if (ioctl(sock, SIOCDELRT, &rt)<0) {
    syslog(LOG_ERR,"Failed to delete route to slan server.\n");
  }

  /* Shut down the slan0 interface here? */
  strcpy(ifr.ifr_name,"slan0");
  ifr.ifr_flags = IFF_NOARP | IFF_BROADCAST;
  if (ioctl(sock, SIOCSIFFLAGS, &ifr) < 0) {
    syslog(LOG_ERR,"Failed to shutdown slan0 interface. \n");
  }

  close(sock);
  if (!dont_restore) {
    syslog(LOG_INFO,"Slan routing deleted, normal routing restored.\n");
  } else {
    syslog(LOG_WARNING,"Default routing was not restored because SLAN "
	   "terminated unexpectedly.\n");
  }
  return;
}

static char *shut_up_compiler(u_int32_t addr) {
  struct in_addr x;
  x.s_addr = addr;

  return inet_ntoa(x);
}

/* This function inspects the system's routing table (as per /proc/net/route)
   and finds the default route. If there are more than one it will
   (incorrectly most likely) use the last one in the table. */
static unsigned int route_table_get_default() {
  unsigned int server_target_route =0x0;
  unsigned char inputline[256];
  FILE *rtable;

  rtable = fopen("/proc/net/route","r");
  if (rtable == NULL) perror("Could not open /proc/net/route.\n");
  while(getc(rtable)!='\n');
  while(!feof(rtable)) {
    unsigned int target, gate, flags;
    fgets(inputline,256,rtable);
    sscanf(inputline,"%*s %x %x %x",&target,&gate,&flags);
    if (target == 0x0 && (flags & 0x1)) {
      server_target_route = gate;
    }
  }

  fclose(rtable);
  return server_target_route;
}


/* On signal SIGUSR1 -- dump status to /var/tmp socket */
static void signal_statusupdate(int signal) {
  int fd;
  FILE *status_file = NULL;

  fd = open(PIPE_FILENAME, O_WRONLY|O_NONBLOCK);
  if (fd < 0) { 
    syslog(LOG_ERR,"Can not open fifo to communicate status information. "
	    "(%s).\n",strerror(errno));
    return;
  } else {
    status_file = fdopen(fd, "w");
  }

  fprintf(status_file,"SLAN is (still) running\n");
  fprintf(status_file,"Server is %s\n",slan_global_config->server_address);
  fprintf(status_file,"Bytes received: %lu kB\tsent: %lu kB\n",
	  session_obj->bytes_in/1024, session_obj->bytes_out/1024);
  fprintf(status_file,"Packets received: %lu\tsent: %lu\n",
	  session_obj->packets_in, session_obj->packets_out);
  fprintf(status_file,"Keys last changed %d seconds ago.\n",
	  time(NULL) - session_obj->key_timestamp);
  fprintf(status_file,"---- END OF STATUS REPORT ----\n");
  fflush(status_file);
  fclose(status_file);
}


/* This function attempts to find the PID of a running SLAN client, if one
   exists, by looking for a file created in /var/run which has the PID of
   a running SLAN client, and checking to see if /proc reports that the 
   name of the program running with that PID (if there is one) is
   slan_client. If all is satisfied, returns the PID of the running SLAN
   client. Note that this will fail to detect a running client if it is run
   under a different name than slan_client. */
static pid_t find_runningpid() {
  pid_t pid, return_value;
  FILE *f;
  char inputline[256];

  f = fopen(PID_FILENAME,"r");
  if (f == NULL) {
    return -1;
  }
  fgets(inputline,255,f);
  inputline[255]=0;
  pid = atol(inputline);
  if (pid < 0 || pid > 0xFFFF) return -1;
  
  fclose(f);
  
  sprintf(inputline,"/proc/%lu/stat",pid);
  f = fopen(inputline,"r");
  if (f == NULL) return -1;
  
  fgets(inputline,255,f);
  inputline[255] = 0;
  if (strstr(inputline, "slan_client")) {
    return_value = pid;
  } else {
    return_value = -1;
  }

  fclose(f);
  return return_value;
}
  

static void client_getstatus() {
  pid_t pid;
  int fd;
  FILE *f;
  fd_set read_fds, err_fds;
  int rv;
  struct timeval timeout;

  timeout.tv_sec = 5;
  timeout.tv_usec = 0;

  pid = find_runningpid();
  if (pid == -1) {
    fprintf(stderr,"SLAN client does not appear to be running.\n");
    exit(-1);
  }

  unlink(PIPE_FILENAME);
  if (mknod(PIPE_FILENAME, (S_IFIFO | 0600), 0)) {
    fprintf(stderr,"Warning: Can not create fifo to communicate status "
	    "information. (%s).\n",strerror(errno));
    return;
  }
  fd = open(PIPE_FILENAME, O_RDONLY | O_NONBLOCK);
  fchown(fd, NOBODY_UID, NOBODY_GID);

  if (fd == -1) {
    fprintf(stderr,"Error: Unable to open SLAN client communication pipe "
	    "for slan client process #%lu (%s)\n",pid,strerror(errno));
    exit(-1);
  }

  if (kill(pid, SIGUSR1)) {
    fprintf(stderr,"Error: Unable to send \"get status\" signal to running "
	    "SLAN client process #%lu (%s)\n",pid,strerror(errno));
    if (errno == EPERM) 
      fprintf(stderr,"Are you root?\n");
    exit(-1);
  }

  f = fdopen(fd, "r");
  FD_ZERO(&read_fds);
  FD_ZERO(&err_fds);
  FD_SET(fd, &read_fds);
  FD_SET(fd, &err_fds);

  rv = select(fd + 1, &read_fds, NULL, &err_fds, &timeout);
  if (rv == -1) {
    fprintf(stderr,"Error: select() failed on slan client pipe. (%s)",
	    strerror(errno));
    exit(-1);
  }
  if (rv == 0) {
    fprintf(stderr,"Time out waiting for running SLAN client (%lu) to "
	    "respond. SLAN client is\nprobably frozen.\n", pid); 
    fprintf(stderr,"Try -k to kill it.\n");
    exit(-1);
  }
  if (FD_ISSET(fd, &err_fds)) {
    fprintf(stderr,"Error: Exception on select() for slan client pipe.\n");
    fprintf(stderr,"Try again or use -k to kill the running slan client.\n");
    exit(-1);
  }
  if (FD_ISSET(fd, &read_fds)) {
    char statusline[1024];
    do {
      fgets(statusline, 1023, f);
      fprintf(stdout,"%s",statusline);
    } while (strncmp(statusline,"----",4) && !feof(f));
    fclose(f);
    unlink(PIPE_FILENAME);
    exit(0);
  }

  fprintf(stderr,"Internal Error. Can not continue (This should not happen)\n");
  exit(0);

}

/* Parse command line options */
static void comline(int argc, char *argv[], char **config_filename)  {

  static struct option long_options[] = {
    { "status", no_argument, NULL, 's' },
    { "kill", no_argument, NULL, 'k' },
    { "config", required_argument, NULL, 'C' },
    { "connect-only", no_argument, NULL, 'r' }
  };

  const char optstring[] = "skC:r";
  pid_t pid;

  opterr = 0;

  while(1) {
    switch(getopt_long(argc, argv, optstring, long_options, NULL)) {
    case -1:
      return;
    case 's':
      client_getstatus();
      break;
    case 'k':
      pid = find_runningpid();
      if (pid > 0) {
	/* Don't do this if we get -1 out of find_runningpid()! -- That will
	   kill everything on the system! */
	if (kill(pid, SIGINT)<0) {
	  fprintf(stderr,"Could not send shutdown signal to slan client "
		  "process %lu (%s)\n",pid,strerror(errno));
	} else {
	  fprintf(stderr,"Sent shutdown signal to SLAN client process %lu\n",
		  pid);
	}
      } else {
	fprintf(stderr,"Can not find running SLAN client.\n");
      }
      exit(0);
      break;
    case 'C':
      *config_filename = strdup(optarg);
      break;
    case 'r':
      connect_only = 1;
      //fprintf(stderr,"Robo-client connect only not implemented.\n");
      break;
    default:
      fprintf(stderr,"Unknown Option %c\n",optopt);
      break;
    }
  }

}

static void client_startup(slan_session_t *session_obj, ip_info_t *ip_conf,
			   int *server_addr, int *server_route) {
  unsigned char null_key[32];
  

  session_obj->control_fd = 
    client_open_socket(slan_global_config->server_address, server_addr, 
		       slan_global_config->control_port);
  session_obj->lock = slan_calloc(1, sizeof(pthread_mutex_t));
  pthread_mutex_init(session_obj->lock, NULL);

  *server_route = route_table_get_default();
  if (*server_route == 0x0) {
    fprintf(stderr,"Error: Cannot find a default route\n");
    fprintf(stderr,"Is your real (physical) network up and configured "
	    "properly?\n");
    exit(1);
  }

  /* Establish whether or not the server is compatible with the client. In the 
     future this may instruct the server (or client) to do fancy things like
     enable or disable certain features or bug fixes. For now, it really
     doesn't do anything. */
  fprintf(stdout,"Checking version compatibility..."); fflush(stdout);
  if (check_version(session_obj->control_fd)<0) {
    fprintf(stderr,"Error, version of server and client incompatible.\n");
    exit(-1);
  } else {
    fprintf(stderr,"Server Version okay.\n");
  }

  fprintf(stdout,"Initializing to default key values...\n"); fflush(stdout);
  /* The DH key exchange will use control message functionality which expects
     to have these things set up and working, except we don't know the key
     until after DH completes. So we'll set them up with null keys (0x0) so
     they function, but they will not provide any security ofcourse. */
  memset(null_key, 0x0, 32);
  session_obj->mac = get_hmac_sha1();
  session_obj->cipher = get_blowfish();
  session_obj->mac->set_key(session_obj->mac, null_key, 32);
  session_obj->cipher->set_key(session_obj->cipher, null_key, 32);
  session_obj->key_timestamp = time(NULL);

  /* Establish the session key via Diffie-Hellman key exchange. The value
     computed by this exchange will be used to produce a symmetric key for
     data privacy and a Message Authentication Code (MAC) key for protecting
     the messages against alterations. We also compute a one-way hash of the
     key value computed and use that to defeat a Diffie-Hellman
     man-in-the-middle attack by using it in the RSA authentication challenge
     which authenticates the server. */
  fprintf(stdout,"Establishing session key with server (Diffie-Hellman)...\n");
  fflush(stdout);
  if (dh_initiate(session_obj)) {
    fprintf(stderr,"Cannot continue, Diffie-Hellman key exchange failed.\n");
    exit(1);
  }

  fprintf(stdout,"Diffie-Hellman Completed. Setting session keys....\n"); 
  fflush(stdout);

  /* Setup the symmetric cipher. */
  session_obj->cipher->set_key(session_obj->cipher, session_obj->key,
			       session_obj->key_length); 

  /* Setup the message authentication code */
  session_obj->mac->set_key(session_obj->mac, session_obj->mackey,
			    session_obj->mackey_length); 
  
  fprintf(stdout,"Symmetric Encryption key: %d bits.\n",
	  session_obj->key_length * 8);
  fprintf(stdout,"Message Authentication Code key: %d bits.\n",
	  session_obj->mackey_length * 8);
  fprintf(stdout,"Finished symmetric session key configuration, link is "
	  "secure.\n\n");
  fflush(stdout);

  /* Authenticate the server. This is a required part of the protocol for any
     service provider. The public key should be given in the service set
     definitions loaded above (some day) */
  fprintf(stdout,"Establishing authenticity of server (RSA)...");
  fflush(stdout);
  if (rsaauth_initiate(session_obj)) { 
    fprintf(stderr,"Server identity cannot be verified: Cannot continue.\n");
    exit(-1);
  }
  fprintf(stdout, "Server identity verified.\n\n");

  fprintf(stdout, "Client (user) authentication...\n");
  if (client_clientauth_wait(session_obj)) {
    fprintf(stderr,"Client authentication to server failed.\n");
    exit(-1);
  }
  fprintf(stdout,"Client (you) accepted.\n\n"); fflush(stdout);

  /* !!!! TEMPORARY CODE !!!!
     
     Let the server dictate to us the IP address it would like to use for
     us. */ 
  fprintf(stdout,"Asking for network configuration parameters...\n");
  fflush(stdout);
  if (ipconf_request(ip_conf, session_obj)) {
    fprintf(stdout,"Could not obtain IP configuration information "
	    "from server.\n");
    slan_memfree(ip_conf);
    exit(-1);
  }

  fprintf(stderr,"Your SLAN session id: %d\n",session_obj->id);
  fprintf(stdout,"Ip address is %s\n",shut_up_compiler(ip_conf->ip_address));
  fprintf(stdout,"Netmask is %s\n",shut_up_compiler(ip_conf->subnet_mask));
  fprintf(stdout,"Broadcast address is %s\n",
	  shut_up_compiler(ip_conf->subnet_bcast));
  fprintf(stdout,"Gateway address is %s\n",
	  shut_up_compiler(ip_conf->router));
  fprintf(stdout,"Nameserver #1 is %s\n",
	  shut_up_compiler(ip_conf->nameserver_1));
  fprintf(stdout,"Nameserver #2 is %s\n\n",
	  shut_up_compiler(ip_conf->nameserver_2)); 
  fflush(stdout);

  client_setup_transport_layer(session_obj,slan_global_config->server_address);

}

static void create_pidfile(void) {
  FILE *f;
  pid_t pid;
  int fd;

  pid = getpid();
  f = fopen(PID_FILENAME, "w");

  if (f == NULL) {
    fprintf(stderr,"Warning: Can not open %s to write process id information. "
	    "(%s)\n",PID_FILENAME,strerror(errno));
    fprintf(stderr,"You will not be able to read status information from the "
	    "running client.\n");
    fprintf(stderr,"SLAN will have to be terminated manually using kill(1)\n");
    fprintf(stderr,"Send the slan_client process with the smallest PID\n");
    fprintf(stderr,"an interrupt signal (SIGINT or kill -2)\nto terminate");
    fprintf(stderr,"your SLAN session.\n");
    return;
  }

  fprintf(f,"%lu\n",pid);
  fclose(f);

  return;
}
	    
    

/* This is the entry point for the client side user space daemon program 

   TO BE COMMENTED MORE LATER */
int main(int argc, char *argv[]) {
  ip_info_t *ip_conf;
  int ipconfig_sock;
  unsigned int server_route, server_addr;
pthread_t transport_thread, control_thread;
  char *config_filename;
  pid_t pid;
  pthread_attr_t thread_attr;

  /* Lock ALL memory -- This process is important to functioning of the
     network for the client, and should never be swapped out. As well (and
     more importantly), the memory used by this process holds vital security
     sensitive information, such as cryptographic keys, and must never be
     written to permanent store. */
  if (mlockall(MCL_FUTURE)) {
    perror("Failed locking memory");
    fprintf(stderr,"This process must be run as root.\n");
    exit(-1);
  }

  /* Before doing anything, gather entropy and setup our random number
     generator. This will be used to provide security, so it better be good. */
  random_initialize();

  /* This function call never returns if command line parsing fails. In 
     fact, it only returns if we are going to start the SLAN client for
     real */
  config_filename = NULL;
  comline(argc, argv, &config_filename);

  if (slan_client_load_configuration(config_filename)) {
    fprintf(stderr,"Cannot start slan: Failed loading configuation file.\n");
    exit(-1);
  }

  pid = find_runningpid();
  if (pid>0) {
    fprintf(stderr,"Appears to be an SLAN client already running (pid %lu)\n",
	    pid);
    fprintf(stderr,"Use --status to try to read its status.\n");
    exit(-1);
  }

  if (!connect_only) {
    struct ifreq ifr;
    strcpy(ifr.ifr_name,"slan0");

    /* Check for slan0 device and take it down if it is running */
    ipconfig_sock = socket(AF_INET, SOCK_DGRAM, 0);
    ifr.ifr_flags = IFF_NOARP | IFF_BROADCAST;
    if (ioctl(ipconfig_sock, SIOCSIFFLAGS, &ifr) < 0) {
      fprintf(stderr,"Can't find interface slan0 (%s)\n",strerror(errno));
      fprintf(stderr,"Is the device loaded (insmod/modprobe)?\n");
      exit(-1);
    }
  } else {
    ipconfig_sock = -1;
  }

  /* Allocate and setup a session object which will hold some state
     information for the active connection. 

     TO BE DONE:
     (1) The server address is to be determined from the service set
     information, when that functionality is added */
  session_obj = slan_calloc(1, sizeof(slan_session_t));
  ip_conf = (ip_info_t *) slan_calloc(1, sizeof(ip_info_t));

  client_startup(session_obj, ip_conf, &server_addr, &server_route);

  if (connect_only) {
close(session_obj->control_fd);
    exit(0);
  }

  fprintf(stdout,"Attempting to configure local network settings "
	  "automatically..."); fflush(stdout);
  if (linux_configure_network(ip_conf, server_route, server_addr, 
			      ipconfig_sock)) {
    fprintf(stdout,"Error, network configuration failed. Cannot continue.\n");
    exit(-1);
  }
  fprintf(stdout,"Done.\n\n"); fflush(stdout);

  slan_memfree(ip_conf);

  /* End of temporary code -- the above will be replaced by something more
     concrete soon. */
  fprintf(stdout,"Session startup okay, SLAN secured network link running.\n");

  /* Lock memory again! Locks are not inherited after fork() */
  if (mlockall(MCL_CURRENT|MCL_FUTURE)) {
    perror("Failed locking memory");
    fprintf(stderr,"This process must be run as root.\n");
    exit(-1);
  }

  pthread_attr_init(&thread_attr);
  pthread_attr_setstacksize(&thread_attr, 65535);
  if (setegid(SLANRUN_GID) || seteuid(SLANRUN_UID)) {
    fprintf(stderr,"SLAN (Warning): Can't change user to drop root "
	    "priveledges\n");
  } else {
    fprintf(stderr,"SLAN client: Root priveledge dropped.\n");
}



  if ((pid=fork())!=0) {
    if (pid == -1) {
      fprintf(stderr,"Can not fork into background! (%s)\n",strerror(errno));
      fprintf(stderr,"Exiting. Something must be way messed up.\n");
      exit(-1);
    } else {
      fprintf(stderr,"Forked into background. Child is [%d]\n",pid);
      return 0;
    }
  } else {
    fclose(stderr);
    fclose(stdout);
    fclose(stdin);
  }


  create_pidfile();
  /* Loop forever until a shutdown event comes or the server stops
     responding. */
  pthread_create(&transport_thread, &thread_attr, transport_loop, session_obj);
  
  pthread_create(&control_thread, &thread_attr, control_loop, session_obj);

  signal(SIGINT, interrupt_shutdown);
  signal(SIGUSR1, signal_statusupdate);
    
  /* When the control thread exits, we are done. Other things may terminate
     the process entirely however.  */
  pthread_join(control_thread, NULL);
  pthread_kill(transport_thread, 15);
  
  if (seteuid(0))
    syslog(LOG_WARNING,"Failed to regain root priveledges (%m)\n");

  client_restore_routes(server_route, server_addr);
  close(session_obj->control_fd);

  syslog(LOG_INFO,"Slan session finished.\n"); fflush(stdout);
  return 0;
}

