/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
  $Header: /home/cvsroot/stage4/platform/windows/winclient/main.c,v 1.4 2001/07/06 17:50:25 lucas Exp $
  $Author: lucas $ $Date: 2001/07/06 17:50:25 $ $Name:  $
  $Revision: 1.4 $
 
*/

#include "slan_windows.h"
#include "slan_types.h"

/* session state type definitions and constants */
#include "slan_session.h"
#include "slan_version.h"
#include "slan_config.h"

/* memory wrapper */
#include "slan_alloc.h"

/* Cryptography */
#include "blowfish.h"
#include "sym_cipher.h"
#include "sha.h"
#include "mac.h"
#include "hmac_sha1.h"

/* Setup and Control Protocol components */
#include "dh_keyexchange.h"
#include "rsaauth.h"
#include "random_interface.h"
#include "client_auth.h"
#include "ip_conf.h"
#include "control_msg.h"
#include "control_tcp.h"
#include "client_pamauth.h"

/* Transport protocol stack */
#include "proto_stack.h"
#include "stack_register.h"

/* slan dll interface */
#include "slandll.h"

/* This variable is global to the entire slan program */
slan_conf_t *slan_global_config;

/* these things are specific to the windows version... */
static WSAEVENT control_read_event = NULL;
static WSAEVENT transport_udp_read_event = NULL;
static void *slan_driver_opaque = NULL;
static HANDLE slan_read_event = NULL;
static HANDLE slan_shutdown_event = NULL;
static HWND slan_parent_window = NULL;
static SOCKET udp_transport_fd; // I don't like this.

/* information required for cleanup */
static BOOL s_changed_routes = FALSE;
static BOOL s_loaded_interface = FALSE;
static BOOL s_started_interface = FALSE;
static BOOL s_configured_interface = FALSE;

static unsigned int g_default_gateway, g_server_addr, g_interface_number;

static void client_setup_transport_layer(slan_session_t *session_obj,
					 char *server_address) {
  proto_stack_t *last, *current;
  int header_save;

  header_save = 0;
  /* A stack terminator at the top -- this is simply a dummy protocol
     layer that drops any packet passed to it. It can be used to catch
     bad or dropped packets and debug them. Otherwise, it isn't necessary and
     doesn't do much. */
  current = last = slan_malloc(sizeof(proto_stack_t));
  terminate_register(current);
  header_save += current->header_reserve;
  
  /* Packets received from the slan device sent down the stack from the top */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  winslan_link_register(current, slan_driver_opaque);
  header_save += current->header_reserve;
  session_obj->transport_top = current;
  current->prev = last;
  last = current;

  /* Optional fragmenting protocol layer -- currently not implemented */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  fragment_register(current);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* All packets being sent through the slan tunnel get tagged with sequence
     numbers before adding a MAC and encrypting. This used to limit the
     possibilities for message replay attacks. */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  sequence_register(current);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* After adding the sequence number, we add a Message Authentication Code
     (MAC) to allow the receiver to verify the integrity of the packet and the 
     sender of the data */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  message_sign_register(current, session_obj->mac);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* Lastly, we enrypt the packet for privacy during transmission. Currently,
     I am expecting the symmetric cipher used to be a configurable option, so
     the cipher used depends on the field in the session_obj */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  symmetric_cipher_register(current, session_obj->cipher);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* Session ID numbers are prepended to identify the session the message
     belongs to */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  session_id_register(current);
  header_save += current->header_reserve;
  *((u_int32_t *) current->data) = session_obj->id;
  current->prev = last;
  last = current;

  /* Transmit the secured packet via the UDP channel */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  winclient_udplink_register(current, server_address,
                            slan_global_config->transport_port,
                            session_obj, &udp_transport_fd);
  header_save += current->header_reserve;
  session_obj->transport_bottom = current;
  current->prev = last;
  last = current;

  /* Terminate the bottom of the stack */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  terminate_register(current);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  session_obj->header_save = header_save;
}

static void forget_keys(slan_session_t *session_obj) {
  int i;

  /* Forget keys. NULL pointers may exist if shutdown happens before the 
     session is configured fully */
  if (session_obj->mackey!=NULL) {
    for(i=0;i<session_obj->mackey_length;i++)
      session_obj->mackey[i] = 0x0;
  }

  /* Usally diffie-hellman returns one big long key which we set the mackey
     pointer to something halfway through and the cipher key to the memory
     returned from DH. So we only free the cipher key. */
  if (session_obj->key!=NULL) {
    for(i=0;i<session_obj->key_length;i++)
      session_obj->key[i] = 0x0;
    slan_memfree(session_obj->key);
  }

}

static void client_shutdown(slan_session_t *session_obj) {
  
  fprintf(stdout,"Connection closing...\n"); fflush(stdout);

  session_obj->cipher->close(session_obj->cipher);
  session_obj->mac->close(session_obj->mac);
  forget_keys(session_obj);

  /* This should cause the process to exit */
  SetEvent(slan_shutdown_event);
}

static void change_key(slan_session_t *session_obj) {

  forget_keys(session_obj);
  if (dh_initiate(session_obj)) {
    fprintf(stderr,"Change session key: DH exchange failed!\n");
    fprintf(stderr,"The key was not changed.\n");
    client_shutdown(session_obj);
  } else {
    fprintf(stdout,"Setting MAC and cipher keys....\n"); fflush(stdout);
    session_obj->mackey_length = session_obj->key_length/2;
    session_obj->key_length -= session_obj->mackey_length;
    session_obj->mackey = session_obj->key + session_obj->key_length;
    
    session_obj->cipher->set_key(session_obj->cipher, session_obj->key,
				 session_obj->key_length); 
    
    session_obj->mac->set_key(session_obj->mac, session_obj->mackey,
			      session_obj->mackey_length); 
    /* Reset sequence numbers: Maybe this should be randomized a bit,
       bleeding off some bits from the DH key exchange. */
    session_obj->sequence_down = 0;
    session_obj->sequence_up = 0;
    
    session_obj->key_timestamp = time(NULL);
    /* Unblock the transport layer. */
  }
  session_obj->key_change = 0;
}

static void client_unexpected_shutdown(slan_session_t *session_obj) 
{
	MessageBox(NULL,
		"The connection to the server has unexpectedly closed.\r\n"
		"The session will be terminated (sorry!).",
		"SLAN Error", MB_OK|MB_ICONSTOP);

  client_shutdown(session_obj);
}

static void main_loop(slan_session_t *session_obj)
{
	control_msg_t *msg, *new_msg;
	u_int32_t type, subtype;
	unsigned int ret;
	int control_channel_only;
	WSAEVENT *event_list[4];
	WSANETWORKEVENTS wsne;
    unsigned char buf[10240];
    unsigned char *p;
    slan_message_t slan_message;

	control_channel_only = 0;

	control_read_event = WSACreateEvent();
	WSAEventSelect(session_obj->control_fd, control_read_event, FD_READ|FD_OOB|FD_CLOSE);

	transport_udp_read_event = WSACreateEvent();
	WSAEventSelect(udp_transport_fd, transport_udp_read_event, FD_READ|FD_OOB|FD_CLOSE);

    slan_message.buffer = buf;
    slan_message.maxspace = 4096;

    p = buf + session_obj->header_save;
  
    slan_message.header = slan_message.data = p;
    slan_message.trailer = p;
    slan_message.length = 0;
    session_obj->down_blocking = session_obj->up_blocking = 0;


	event_list[0] = control_read_event;
	event_list[1] = transport_udp_read_event;
	event_list[2] = (WSAEVENT) slan_read_event;
	event_list[3] = (WSAEVENT) slan_shutdown_event;

    for(;;) /* forever... */
	{
		ret = WSAWaitForMultipleEvents(4, (HANDLE *) event_list, FALSE, 10000, FALSE);
		if(ret == WSA_WAIT_FAILED)
		{
			fprintf(stderr, "WSA_WAIT_FAILED! (what the hell does that mean?)\n");
			continue;
		}

		if(ret == WSA_WAIT_TIMEOUT)
		{
			/* Check the session key's age. Change it if it should be expired. This is
			the client's responsibility to enforce. */
			if ((time(NULL) - session_obj->key_timestamp) > KEY_LIFETIME &&
				session_obj->key_change == 0) 
			{
				session_obj->key_change = 1;
				new_msg = new_control_msg(KEYCHANGE, KEYCHANGE_REQUEST, session_obj);
				control_msg_send(new_msg);
				control_msg_destruct(new_msg);
			}

			fprintf(stderr, "zzzz...\n");
			continue;
		}

		if(ret == WSA_WAIT_EVENT_0 + 3)
		{
			fprintf(stderr, "user request -- shutting down...\n");
			client_shutdown(session_obj);
			return;
		}

		if(ret == WSA_WAIT_EVENT_0)
		{
			ret = WSAEnumNetworkEvents(session_obj->control_fd, control_read_event, &wsne);
			if(ret)
			{
				fprintf(stderr, "WSAEnumNetworkEvents failed!\n");
				continue;
			}

			if(wsne.lNetworkEvents & FD_OOB)
			{
				fprintf(stderr,"Out of band data waiting... why is that?\n");
				continue;
			}
			else if(wsne.lNetworkEvents & FD_CLOSE)
			{
                /* If the server hangs up, we shutdown immediately. */
                client_unexpected_shutdown(session_obj);
				return;
			}
			else if(wsne.lNetworkEvents & FD_READ)
			{
				/* Yay, we have data! */

				msg = control_msg_receive(session_obj, &type, &subtype, 0);
				if(msg == NULL)
					continue;

				switch(type) 
				{
					case KEYCHANGE:
						if (subtype == KEYCHANGE_READY) 
						{
							change_key(session_obj);
							control_msg_destruct(msg);
						} 
						else 
						{
							session_obj->key_change = 1;
							new_msg = new_control_msg(KEYCHANGE, KEYCHANGE_READY, session_obj);
							control_msg_send(new_msg);
							control_msg_destruct(new_msg);
							control_msg_destruct(msg);
							// usleep(1000); // what the hell is this??
							change_key(session_obj);
						}
						continue;
						break;
      
					case SHUTDOWN:
						client_shutdown(session_obj);
						return;
      
					case KEEPALIVE:
						if (subtype != KEEPALIVE_PING) 
						{
							/* This should not happen. */
							fprintf(stderr,"Not expecting a keepalive response here.\n");
						} 
						else 
						{
							new_msg = new_control_msg(KEEPALIVE, KEEPALIVE_RESPONSE, session_obj);
							control_msg_send(new_msg);
							control_msg_destruct(new_msg);
							fprintf(stderr,"Sent keepalive response.\n");
						}
						break;
      
					default:
						fprintf(stderr,"Unhandled control message type: %d\n",type);
						break;
				}
				control_msg_destruct(msg);
			}
			continue;
		}

		if(ret == WSA_WAIT_EVENT_0 + 1) /* transport udp read event.. */
		{
			WSAResetEvent(transport_udp_read_event);
			session_obj->transport_bottom->up(session_obj->transport_bottom,
					&slan_message);
			slan_message.header = slan_message.data = p;
			slan_message.trailer = p;
			slan_message.length = 0;

			continue;
		}

		if(ret == WSA_WAIT_EVENT_0 + 2  && !session_obj->key_change)
		{
			session_obj->transport_top->down(session_obj->transport_top,
					 &slan_message);
			slan_message.header = slan_message.data = p;
			slan_message.trailer = p;
			slan_message.length = 0;

			ResetEvent(slan_read_event);
			continue;
		}
	}
}

static char *shut_up_compiler(u_int32_t addr) {
  struct in_addr x;
  x.s_addr = addr;

  return inet_ntoa(x);
}

static int route_add(unsigned int ip, unsigned int mask, unsigned int gw)
{
  char windir[256];     // path to winipcfg.exe
  char tmp[32];			// temporary number.
  unsigned int ret;     // return value (generic)
  STARTUPINFO sui;      // for createprocess
  PROCESS_INFORMATION pinfo; // ''

  ret = GetWindowsDirectory(windir, 128);
  if(!ret || ret > 128)
  {
	  MessageBox(NULL, "Could not locate the windows directory.", "Fatal Error", MB_OK|MB_ICONSTOP);
	  return -1;
  }

  if(windir[ret] != '\\')
  {
	  windir[ret++] = '\\';
	  windir[ret] = '\0';
  }

  strcat(windir, "ROUTE.EXE ");
  sprintf(tmp, "add %s ", shut_up_compiler(ip));
  strcat(windir, tmp);
  sprintf(tmp, "mask %s ", shut_up_compiler(mask));
  strcat(windir, tmp);
  sprintf(tmp, "%s", shut_up_compiler(gw));
  strcat(windir, tmp);

  memset(&sui, 0, sizeof(STARTUPINFO));
  sui.cb = sizeof(STARTUPINFO);

  if(CreateProcess(NULL, windir, NULL, NULL, FALSE, 
	  DETACHED_PROCESS, NULL, "C:\\WINDOWS", &sui, &pinfo) == 0)
  {
	  return -1;
  }

  WaitForSingleObject(pinfo.hProcess, INFINITE);

  CloseHandle(pinfo.hProcess);
  CloseHandle(pinfo.hThread);

  return 0;
}

static int route_delete(unsigned int ip)
{
  char windir[256];     // path to winipcfg.exe
  char tmp[32];			// temporary number.
  unsigned int ret;     // return value (generic)
  STARTUPINFO sui;      // for createprocess
  PROCESS_INFORMATION pinfo; // ''

  ret = GetWindowsDirectory(windir, 128);
  if(!ret || ret > 128)
  {
	  MessageBox(NULL, "Could not locate the windows directory.", "Fatal Error", MB_OK|MB_ICONSTOP);
	  return -1;
  }

  if(windir[ret] != '\\')
  {
	  windir[ret++] = '\\';
	  windir[ret] = '\0';
  }

  strcat(windir, "ROUTE.EXE ");
  sprintf(tmp, "delete %s", shut_up_compiler(ip));
  strcat(windir, tmp);

  memset(&sui, 0, sizeof(STARTUPINFO));
  sui.cb = sizeof(STARTUPINFO);

  if(CreateProcess(NULL, windir, NULL, NULL, FALSE, 
	  DETACHED_PROCESS, NULL, "C:\\WINDOWS", &sui, &pinfo) == 0)
  {
	  return -1;
  }

  WaitForSingleObject(pinfo.hProcess, INFINITE);

  CloseHandle(pinfo.hProcess);
  CloseHandle(pinfo.hThread);

  return 0;
}


static int windows_deconfigure_network(int ifnumber)
{
  char windir[256];     // path to winipcfg.exe
  char tmp[16];			// temporary number.
  unsigned int ret;     // return value (generic)
  STARTUPINFO sui;      // for createprocess
  PROCESS_INFORMATION pinfo; // ''

  ret = GetWindowsDirectory(windir, 128);
  if(!ret || ret > 128)
  {
	  MessageBox(NULL, "Could not locate the windows directory.", "Fatal Error", MB_OK|MB_ICONSTOP);
	  return -1;
  }

  if(windir[ret] != '\\')
  {
	  windir[ret++] = '\\';
	  windir[ret] = '\0';
  }

  strcat(windir, "WINIPCFG.EXE /batch WINSLAN.OUT /release ");
  sprintf(tmp, "%d", ifnumber);
  strcat(windir, tmp);

  memset(&sui, 0, sizeof(STARTUPINFO));
  sui.cb = sizeof(STARTUPINFO);

  if(CreateProcess(NULL, windir, NULL, NULL, FALSE, 
	  DETACHED_PROCESS, NULL, "C:\\WINDOWS", &sui, &pinfo) == 0)
  {
	  return -1;
  }

  WaitForSingleObject(pinfo.hProcess, INFINITE);

  _unlink("WINSLAN.OUT");

  CloseHandle(pinfo.hProcess);
  CloseHandle(pinfo.hThread);

  return 0;
}


static int windows_configure_network(ip_info_t *ip_conf, int ifnumber)
{ 
  struct local_info li; // local info to pass to slan dll
  char windir[256];     // path to winipcfg.exe
  char tmp[16];			// temporary number.
  unsigned int ret;     // return value (generic)
  STARTUPINFO sui;      // for createprocess
  PROCESS_INFORMATION pinfo; // ''
  BOOL done = FALSE;    // to break out of waiting loop
  HANDLE *harray[3];    // ..
  char buffer[65536];   // for bogus reads
  int bufferlength = 65536;

  memset(&li, 0, sizeof(struct local_info));

  li.ip = ip_conf->ip_address;
  li.netmask = ip_conf->subnet_mask;
  li.gateway = ip_conf->router;
  li.ns1 = ip_conf->nameserver_1;
  li.ns2 = ip_conf->nameserver_2;

  set_slan_info(slan_driver_opaque, &li);

  ret = GetWindowsDirectory(windir, 128);
  if(!ret || ret > 128)
  {
	  MessageBox(NULL, "Could not locate the windows directory.", "Fatal Error", MB_OK|MB_ICONSTOP);
	  return -1;
  }

  if(windir[ret] != '\\')
  {
	  windir[ret++] = '\\';
	  windir[ret] = '\0';
  }

  strcat(windir, "WINIPCFG.EXE /batch WINSLAN.OUT /renew ");
  sprintf(tmp, "%d", ifnumber);
  strcat(windir, tmp);

  memset(&sui, 0, sizeof(STARTUPINFO));
  sui.cb = sizeof(STARTUPINFO);

  if(CreateProcess(NULL, windir, NULL, NULL, FALSE, 
	  DETACHED_PROCESS, NULL, "C:\\WINDOWS", &sui, &pinfo) == 0)
  {
	  MessageBox(NULL, "Error spawning WinIpCfg", "Fatal Error", MB_OK|MB_ICONSTOP);
	  return -1;
  }

  harray[0] = slan_read_event;
  harray[1] = pinfo.hProcess;
  harray[2] = slan_shutdown_event;

  while (!done)
  {
	  ret = WaitForMultipleObjects(2, (const HANDLE *) harray, FALSE, 1000);

	  if(ret == WAIT_TIMEOUT)
	  {
		  fprintf(stderr, "still waiting for winipcfg...\n");
		  continue;
	  }

  	  if(ret == WAIT_OBJECT_0 + 2)
	  {
		  TerminateProcess(pinfo.hProcess, 0);
		  _unlink("WINSLAN.OUT");

		  CloseHandle(pinfo.hProcess);
	      CloseHandle(pinfo.hThread);

		  return -1;
	  }

	  if(ret == WAIT_OBJECT_0)
	  {
		  bufferlength = 65536;
		  if(!read_slan_interface(slan_driver_opaque, buffer, &bufferlength, FALSE))
		  {
			  fprintf(stderr, "Error reading slan packet!\n");
			  continue;
		  }
		  else
		  {
			  char *mainbuf = buffer;
			  DWORD mainbuflen = bufferlength;
			  char *curbuf;
			  DWORD curlen;
			  BOOL cont;

			  while(1)
			  {
				  cont = get_slan_packet(slan_driver_opaque, &mainbuf, &mainbuflen, &curbuf, &curlen);
				  printf("packet, %d bytes\n", curlen);

		          // mainbuflen is -1 if there was an internal error. This is a bad thing (tm)
		          // and should never really happen.

		          if(mainbuflen == 0xFFFFFFFF)
			         break;

		          // the current 'packet' could have been processed by the DLL layer.
		          // If it was, ignore it. 

		          if(curlen == 0)
				  {
					  if(cont)
						  continue;
				  }
		          else
				  {
					  printf("packet of useless information...\n");
				  }
	
			      if(!cont)
					  break;	
			  }
	
  			  if(mainbuflen == 0xFFFFFFFF)
			  {
				  fprintf(stderr, "Strange error...\n");
			  }
			  continue;
		  }
	  }

	  if(ret == WAIT_OBJECT_0 + 1 || ret == WAIT_ABANDONED_0 + 1)
	  {
		  done = TRUE;
		  continue;
	  }
  }

  _unlink("WINSLAN.OUT");

  CloseHandle(pinfo.hProcess);
  CloseHandle(pinfo.hThread);

  return 0;
}

void slan_stop(int complain)
{
	if(s_changed_routes)
	{
		route_add(0, 0, g_default_gateway);
		route_delete(g_server_addr);
		s_changed_routes = FALSE;
	}

	if(s_configured_interface)
	{
		windows_deconfigure_network(g_interface_number);
		s_configured_interface = FALSE;
	}

	if(s_started_interface)
	{
		stop_slan_interface(slan_driver_opaque);
		s_started_interface = FALSE;
	}

	if(s_loaded_interface)
	{
		close_slan_interface(slan_driver_opaque);
		s_loaded_interface = FALSE;
	}

	memory_cleanup();
	mem_destroy();

	stop_all_gui();

	ExitThread(0);
}

int slan_start(slan_thread_start_t *threadinfo) 
{
  slan_session_t *session_obj;
  unsigned char null_key[32];
  ip_info_t *ip_conf;
  unsigned int server_addr;
  unsigned long err;
  unsigned int default_gateway = 0;
  int interface_number = 0;

  slan_shutdown_event = threadinfo->slan_die_event;

  if(!mem_init())
  {
	  MessageBox(NULL, "Could not initialize memory! Cannot continue!", "SLAN Fatal Error",
		  MB_OK|MB_ICONSTOP);
	  return -1;
  }

  /* try to load the driver. */
  new_slan_status(0, "Loading SLAN driver...", FALSE);

  slan_driver_opaque = load_slan_interface(&err);
  if(slan_driver_opaque == NULL)
  {
	  char tmp[64];

	  sprintf(tmp, "Error loading SLAN interface: %d\n", err);
	  MessageBox(NULL, tmp, "Fatal Error", MB_OK|MB_ICONSTOP);
	  return -1;
  }

  s_loaded_interface = TRUE;

  new_slan_status(0xFFFFFFFF, "Initializing Entropy...", FALSE);

  /* Before doing anything, gather entropy and setup our random number
     generator. This will be used to provide security, so it better be good. */
  random_initialize();

  new_slan_status(0xFFFFFFFF, "Loading Configuration", FALSE);

  if (slan_client_load_configuration(threadinfo->slan_server_config)) 
  {
	  MessageBox(NULL, "Error loading specified configuration file.", "Fatal Error",
		  MB_OK|MB_ICONSTOP);
	  return -1;
  }

  new_slan_status(0xFFFFFFFF, "Getting default route...", FALSE);
  if(get_default_gateway(&default_gateway, &interface_number))
  {
	  MessageBox(NULL, "Could not get default route!", "Fatal Error",
		  MB_OK|MB_ICONSTOP);
	  return -1;
  }

  session_obj = slan_calloc(1, sizeof(slan_session_t));

  new_slan_status(1, "Connecting to server...", FALSE);

  session_obj->control_fd = client_open_socket(slan_global_config->server_address, &server_addr, slan_global_config->control_port);

  new_slan_status(0xFFFFFFFF, "Setting new route...", FALSE);

  if(default_gateway != 0)
  {
	route_add(server_addr, 0xFFFFFFFF, default_gateway);
	route_delete(0);
	g_default_gateway = default_gateway;
	g_server_addr = server_addr;
	s_changed_routes = TRUE;
  }

//  server_route = route_table_get_default();
//  if (server_route == 0x0) {
//    fprintf(stderr,"Error: Cannot find a default route\n");
//    exit(1);
//  }

  /* Establish whether or not the server is compatible with the client. In the 
     future this may instruct the server (or client) to do fancy things like
     enable or disable certain features or bug fixes. For now, it really
     doesn't do anything. */
  new_slan_status(0xFFFFFFFF, "Verifying server version compatibility...", FALSE);

  fprintf(stdout,"Checking version compatibility..."); fflush(stdout);
  if (check_version(session_obj->control_fd)<0) {
    MessageBox(NULL, "Error, version of server and client incompatible.",
		"SLAN Fatal Error", MB_OK|MB_ICONSTOP);
    return -1;
  } else {
    fprintf(stderr,"Server Version okay.\n");
  }

  new_slan_status(0xFFFFFFFF, "Server version okay.", FALSE);

  fprintf(stdout,"Initializing to default key values...\n"); fflush(stdout);
  /* The DH key exchange will use control message functionality which expects
     to have these things set up and working, except we don't know the key
     until after DH completes. So we'll set them up with null keys (0x0) so
     they function, but they will not provide any security ofcourse. */
  memset(null_key, 0x0, 32);
  session_obj->mac = get_hmac_sha1();
  session_obj->cipher = get_blowfish();
  session_obj->mac->set_key(session_obj->mac, null_key, 32);
  session_obj->cipher->set_key(session_obj->cipher, null_key, 32);
  session_obj->key_timestamp = time(NULL);

  /* Establish the session key via Diffie-Hellman key exchange. The value
     computed by this exchange will be used to produce a symmetric key for
     data privacy and a Message Authentication Code (MAC) key for protecting
     the messages against alterations. We also compute a one-way hash of the
     key value computed and use that to defeat a Diffie-Hellman
     man-in-the-middle attack by using it in the RSA authentication challenge
     which authenticates the server. */

  new_slan_status(2, "Establishing session key... (DH)", FALSE);

  fprintf(stdout,"Establishing session key with server (Diffie-Hellman)...\n");
  fflush(stdout);
  if (dh_initiate(session_obj)) {
    fprintf(stderr,"Cannot continue, Diffie-Hellman key exchange failed.\n");
	MessageBox(NULL, "Key exchange with server failed", "SLAN Fatal Error",
		MB_OK|MB_ICONSTOP);
    return -1;
  }

  new_slan_status(0xFFFFFFFF, "Key exchange completed. Setting up session keys...", FALSE);

  fprintf(stdout,"Diffie-Hellman Completed. Setting session keys....\n"); 
  fflush(stdout);

  /* Setup the symmetric cipher. */
  session_obj->cipher->set_key(session_obj->cipher, session_obj->key,
			       session_obj->key_length); 

  /* Setup the message authentication code */
  session_obj->mac->set_key(session_obj->mac, session_obj->mackey,
			    session_obj->mackey_length); 
  
  fprintf(stdout,"Symmetric Encryption key: %d bits.\n",
	  session_obj->key_length * 8);
  fprintf(stdout,"Message Authentication Code key: %d bits.\n",
	  session_obj->mackey_length * 8);
  fprintf(stdout,"Finished symmetric session key configuration, link is "
	  "secure.\n\n");
  fflush(stdout);

  /* Authenticate the server. This is a required part of the protocol for any
     service provider. The public key should be given in the service set
     definitions loaded above (some day) */

  new_slan_status(3, "Verifying authenticity of server...", FALSE);

  fprintf(stdout,"Establishing authenticity of server (RSA)...");
  fflush(stdout);
  if (rsaauth_initiate(session_obj)) { 
    fprintf(stderr,"Server identity cannot be verified: Cannot continue.\n");
	MessageBox(NULL, "Could not verify server authenticity.", "SLAN Fatal Error",
		MB_OK|MB_ICONSTOP);
    return -1;
  }
  fprintf(stdout, "Server identity verified.\n\n");

  new_slan_status(0xFFFFFFFF, "Server identity verified.", FALSE);

  new_slan_status(4, "Authenticating user...", FALSE);
  fprintf(stdout, "Client (user) authentication...\n");
  if (client_clientauth_wait(session_obj)) {
	MessageBox(NULL, "Authentication failed.", "SLAN Fatal Error",
		MB_OK|MB_ICONSTOP);
    fprintf(stderr,"Client authentication to server failed.\n");
    return -1;
  }
  fprintf(stdout,"Client (you) accepted.\n\n"); fflush(stdout);

  new_slan_status(0xFFFFFFFF, "Asking for network configuration parameters...", FALSE);

  fprintf(stdout,"Asking for network configuration parameters...\n");
  fflush(stdout);

  ip_conf = (ip_info_t *) slan_calloc(1, sizeof(ip_info_t));
  if (ipconf_request(ip_conf, session_obj)) {
	MessageBox(NULL, "Could not obtain IP configuration information", "SLAN Fatal Error",
		MB_OK|MB_ICONSTOP);
    fprintf(stdout,"Could not obtain IP configuration information "
	    "from server.\n");
    slan_memfree(ip_conf);
    return -1;
  }

  new_slan_status(5, "Configuring network settings...", FALSE);

  /* create our read event -- this is triggered when data is waiting */
  slan_read_event = CreateEvent(NULL, FALSE, FALSE, NULL);

  /* start the interface (whoa!) */
  if(!slan_read_event || start_slan_interface(slan_driver_opaque, slan_read_event) == FALSE)
  {
	  if(slan_read_event)
		  CloseHandle(slan_read_event);

	  MessageBox(NULL, "Could not start SLAN interface. Strange.", "Fatal Error", MB_OK|MB_ICONSTOP);
	  return -1;
  }
  
  s_started_interface = TRUE;

  fprintf(stdout,"Ip address is %s\n",shut_up_compiler(ip_conf->ip_address));
  fprintf(stdout,"Netmask is %s\n",shut_up_compiler(ip_conf->subnet_mask));
  fprintf(stdout,"Broadcast address is %s\n",shut_up_compiler(ip_conf->subnet_bcast));
  fprintf(stdout,"Gateway address is %s\n",shut_up_compiler(ip_conf->router));
  fprintf(stdout,"Nameserver #1 is %s\n",shut_up_compiler(ip_conf->nameserver_1));
  fprintf(stdout,"Nameserver #2 is %s\n\n",shut_up_compiler(ip_conf->nameserver_2)); 
  fflush(stdout);

  fprintf(stdout,"Attempting to configure local network settings "
	  "automatically..."); fflush(stdout);

  /* sheesh, this is grosser than I thought... */
  if(windows_configure_network(ip_conf, interface_number))
	  return -1;

  g_interface_number = interface_number;
  s_configured_interface = TRUE;

  new_slan_status(5, "Establishing internal transport layer...", FALSE);

  fprintf(stdout,"Done.\n\n"); fflush(stdout);

  slan_memfree(ip_conf);

  fprintf(stdout,"Session startup okay, SLAN secured network link running.\n");
  client_setup_transport_layer(session_obj, slan_global_config->server_address);

  new_slan_status(0xFFFFFFFF, "SLAN secured network link running.", FALSE);

  new_slan_status(0xFFFFFFFF, 0, TRUE);

  SetConsoleTitle("SLAN Debug Window");

  main_loop(session_obj);

  closesocket(session_obj->control_fd);
  closesocket(udp_transport_fd);

  if(default_gateway != 0)
  {
	route_add(0, 0, default_gateway);
	route_delete(server_addr);
	s_changed_routes = FALSE;
  }

  windows_deconfigure_network(interface_number);
  s_configured_interface = FALSE;

  fprintf(stdout,"Slan session finished.\n"); fflush(stdout);
  return 0;
}

DWORD WINAPI slan_thread_wrapper(LPVOID param)
{
	slan_stop(slan_start((slan_thread_start_t *) param));
	return 0;
}

HANDLE start_slan_thread(char *config_file, HANDLE dieevent)
{
	DWORD dwid;
	static slan_thread_start_t threadinfo;

	threadinfo.slan_server_config = config_file;
	threadinfo.slan_die_event = dieevent;

	return CreateThread(NULL, 0, slan_thread_wrapper, &threadinfo, 0, &dwid);
}

