/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
  $Header: /home/cvsroot/stage4/platform/windows/winclient/winclient_udplink.c,v 1.4 2001/07/06 17:50:25 lucas Exp $
  $Author: lucas $ $Date: 2001/07/06 17:50:25 $ $Name:  $
  $Revision: 1.4 $
 
*/

#include "slan_windows.h"

#include "slan_alloc.h"
#include "slan_types.h"

#include "proto_stack.h"
#include "slan_version.h"
#include "slan_session.h"

typedef struct {
  SOCKET fd;
  struct sockaddr_in server_addr;
  slan_session_t *session_obj;
} udplink_t;

/* server_open_socket()

   This function opens a socket to the kernel's UDP/IP layer, for sending 
   packets encapsulated packets to the server.

   TODO: Should this try to bind a specific interface and port? If not, it
   will require that the server learn the port and address to use. */
static void server_open_socket(udplink_t *local) {
  SOCKET sock;
  struct protoent *protocol;
  struct sockaddr_in address;
  
  protocol = getprotobyname("udp");
  if (protocol == NULL) {
    fprintf(stderr,"sland: A fatal error has occured, cannot find the "
	    "operating system protocol number for UDP.\n");
    slan_stop(-1);
  }

  sock = socket(PF_INET, SOCK_DGRAM, protocol->p_proto);
  if (sock == -1 ) {
    perror("sland: Cannot open network socket for UDP/IP: ");
    slan_stop(-1);
  }

  address.sin_family = AF_INET;
  address.sin_addr.s_addr = htonl(INADDR_ANY);
  address.sin_port = 0;

  if (bind(sock, (struct sockaddr *) &address, sizeof(struct sockaddr_in))) {
    perror("sland: Cannot bind UDP/IP socket: ");
    slan_stop(-1);
  }

  fprintf(stderr,"UDP/IP socket created.\n");
  local->fd = sock;
}

/* udplink_set_server()

   Attempts to resolve the string containing the server's hostname to an
   IP address. If it was specified as an IP address, gethostbyname() should
   convert it for us to network byte order binary. Then we setup a sockaddr
   struct to use when pushing datagrams to the server. server_addr is a
   file scope variable, statically allocated. */
static void udplink_set_server(udplink_t *local, char *server_address, int port) {
  struct hostent *hostname_hent;

  if((local->server_addr.sin_addr.s_addr = inet_addr(server_address)) == INADDR_NONE)
  {
    hostname_hent = gethostbyname(server_address);
    if (hostname_hent == NULL) {
#ifndef WIN32
      fprintf(stderr,"client: Could not resolve server hostname %s: %s\n",
              server_name, hstrerror(h_errno));
#endif
      slan_stop(-1);
    }

    memcpy(&local->server_addr.sin_addr.s_addr,hostname_hent->h_addr_list[0],
	 hostname_hent->h_length);
  }

  local->server_addr.sin_port = htons(port);
  local->server_addr.sin_family = AF_INET;
}

static void client_udplink_down(proto_stack_t *stack, slan_message_t
				*message) {
  udplink_t *local;
  int len;

  local = stack->data;
  len = sendto(local->fd,message->data,message->length, 0,
	       (struct sockaddr *) &local->server_addr, sizeof(struct sockaddr_in));
  if (len < message->length) {
    perror("sland: Error sending packet to server: ");
  }
  
  stack->next->down(stack->next,message);
}

static void client_udplink_up(proto_stack_t *stack, slan_message_t
			      *message) {
  udplink_t *local;
  int s_length, len;
  struct sockaddr_in source_addr;

  local = stack->data;

  s_length = sizeof(source_addr);
  len = recvfrom(local->fd, message->data, message->maxspace, 0,
 		 (struct sockaddr *) &source_addr, &s_length); 
  if (len < 0) {
    /* Debugging */
    perror("sland: error receiving data from server: ");
    slan_free(message);
    return;
  }
  if (memcmp(&source_addr.sin_addr.s_addr,
	     &local->server_addr.sin_addr.s_addr,
	     sizeof(source_addr.sin_addr))) {
    /* Debugging */
    fprintf(stderr,"While waiting for the server, we got a packet from "
	    "someplace else!\n");
    fprintf(stderr,"Source of packet reported by network layer is %s\n",
	    inet_ntoa(source_addr.sin_addr));
    slan_free(message);
    return;
  }

  message->length = len;
  stack->prev->up(stack->prev, message);
}


int winclient_udplink_register(proto_stack_t *stack_entry, char *address,
                              int port, slan_session_t *session_obj,
                              SOCKET *udp_fd) {
  udplink_t *local;

  stack_entry->up = client_udplink_up;
  stack_entry->down = client_udplink_down;
  stack_entry->header_reserve = 0;
  stack_entry->trailer_reserve = 0;
  
  local = (udplink_t *) slan_calloc(1, sizeof(udplink_t));
  if (local == NULL) return 1;
  
  server_open_socket(local);
  udplink_set_server(local, address, port);

  local->session_obj = session_obj;

  stack_entry->data = local;

  if(udp_fd)
	  *udp_fd = local->fd;
  
  return 0;
}
