/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage4/transport/slan_link.c,v 1.10 2001/07/07 23:13:58 koni Exp $ 
  $Author: koni $ $Date: 2001/07/07 23:13:58 $ $Name:  $
  $Revision: 1.10 $

*/
   
/*

  KONI Oct. 28-29, 2000
  Stage 1 _client_ daemon process for "lightlink secure lan" project
  
  slan_link.c : This file implements functions to open, send, and receive data
                from the slan device driver running on the local machine.

  Oct 29: File split off from UDP/IP communications because of header file
          conflicts.

*/

/* ---- Include files ----- */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <linux/if_packet.h>
#include <linux/ip.h>
#include <unistd.h>
#include <syslog.h>
#include <sys/ioctl.h>
#include <linux/if_arp.h>
#include <sys/poll.h>
#include "proto_stack.h"

#include "slan_version.h"
#include "slan_session.h"

#include "slan_alloc.h"

/* ARG! HACK */
#define ARPHRD_VOID 0xFFFF

typedef struct {
  int fd;
  struct sockaddr_ll pl_addr;
  slan_session_t *session_obj;
} slan_link_t;

/* open_slan_socket()

   This function opens a network socket of type PF_PACKET, which allows us to
   send frames of data directly to a network device driver, or receive frames
   from a network device driver. It also attempts to find the network device
   index number of our device driver, slan0. If it cannot open the socket or
   it cannot find the index number of slan0, we abort the program. */
static int slan_open_socket() {
  int fd;  

  fd = socket(PF_PACKET, SOCK_RAW, ETH_P_SLAN);
  if (fd < 0) {
    perror("sland: Error opening socket to slan device: ");
    syslog(LOG_CRIT,"Failed to open socket to SLAN kernel driver.\n");
    exit(-1);
  }
  
  return fd;
}

/* This gets a message from the slan device. If a message is
   available, it sends it through the protocol stack. If a message is
   not available, it simply exits the protocol stack. */
static void slan_link_down(proto_stack_t *stack, slan_message_t
			   *message) {
  slan_link_t *local;
  fd_set rfds;
  struct timeval tv;
  int len, retval;

  local = (slan_link_t *) stack->data;
  
  len = read(local->fd, message->data, message->maxspace);
  if (len < 0) {
    /* Debugging stuff */
    syslog(LOG_ERR,"Error reading data from SLAN device socket (%m)\n");
    slan_free(message);
    return;
  }

  message->length = len;

  /* Remove header added by slan device */
  slan_pop(message, sizeof(slan_header_t));

  stack->next->down(stack->next, message);
}

static void slan_link_up(proto_stack_t *stack, slan_message_t
			 *message) {
  slan_link_t *local;
  slan_header_t *header;
  int len;

  local = (slan_link_t *) stack->data;

  slan_reserve(message, sizeof(slan_header_t));
  header = (slan_header_t *) message->header;

  header->magic_number = (0xFF00FF00);
  header->our_protocol = htons(ETH_P_SLAN);
  header->orig_protocol = htons(ETH_P_IP);

  slan_set(message);
  
  len = sendto(local->fd, message->data, message->length, 0, 
	       (struct sockaddr *) &local->pl_addr,sizeof(struct sockaddr_ll));
#ifdef DEBUG
  fprintf(stderr,"slan_link: Packet sent to slan device.\n");
#endif
  if (len < message->length) {
    /* Debugging */
    syslog(LOG_ERR,"Failed to send a packet to SLAN device driver (%m)\n");
    return;
  }

  local->session_obj->bytes_in += message->length;
  local->session_obj->packets_in ++;
  stack->prev->up(stack->prev, message);
}

int slan_link_register(proto_stack_t *stack_entry, int slan_socket, 
		       slan_session_t *session_obj) {
  slan_link_t *local;
  struct ifreq request;
  int slan_index;

  stack_entry->up = slan_link_up;
  stack_entry->down = slan_link_down;
  stack_entry->header_reserve = sizeof(slan_header_t);
  stack_entry->trailer_reserve = 0;

  local = (slan_link_t *) slan_calloc(1, sizeof(slan_link_t));
  if (local == NULL) return -1;

  if (slan_socket == 0) {
    local->fd = slan_open_socket();
  } else {
    local->fd = slan_socket;
  }

  memset(&request,0x0,sizeof(request));
  strcpy(request.ifr_name,"slan0");
  ioctl(local->fd, SIOCGIFINDEX, &request);

  slan_index = request.ifr_ifindex;
  if (slan_index==0) {
    fprintf(stderr,"Cannot find slan0 device driver. Did you insmod it?\n");
    syslog(LOG_CRIT,"Cannot find slan device driver. Not loaded (insmod)?\n");
    exit(-1);
  }

  /* Setup address structure so we can tell the kernel to send packets to 
     slan interface instead of any old interface. This only needs to be once,
     the key parameter is the slan interface index. 

     TODO: I don't know what would happen if the user were to deconfigure and
     remove the slan module, and then reinstall it and reconfigure it. This
     would change the interface index, but, if this program was not also 
     restarted, we'd probably get into trouble. */
  local->pl_addr.sll_family = AF_PACKET;
  local->pl_addr.sll_protocol = 0x7777;
  local->pl_addr.sll_ifindex = slan_index;
  local->pl_addr.sll_hatype = ARPHRD_VOID;
  local->pl_addr.sll_pkttype = PACKET_BROADCAST;
  local->pl_addr.sll_halen = 0;
  memset(local->pl_addr.sll_addr,0x0,8);

  local->session_obj = session_obj;

  stack_entry->data = local;
  session_obj->slan_socket = local->fd;

  return 0;
}

